from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any, Literal

import pydantic as pd
from pydantic_extra_types.semantic_version import SemanticVersion  # noqa: TCH002

from hugr.hugr.base import Hugr
from hugr.utils import deser_it

from .ops import Value
from .serial_hugr import SerialHugr, serialization_version
from .tys import (
    ConfiguredBaseModel,
    ExtensionId,
    ExtensionSet,
    PolyFuncType,
    TypeBound,
    TypeParam,
)

if TYPE_CHECKING:
    from .ops import Value
    from .serial_hugr import SerialHugr


class ExplicitBound(ConfiguredBaseModel):
    b: Literal["Explicit"] = "Explicit"
    bound: TypeBound

    def deserialize(self) -> ext.ExplicitBound:
        return ext.ExplicitBound(bound=self.bound)


class FromParamsBound(ConfiguredBaseModel):
    b: Literal["FromParams"] = "FromParams"
    indices: list[int]

    def deserialize(self) -> ext.FromParamsBound:
        return ext.FromParamsBound(indices=self.indices)


class TypeDefBound(pd.RootModel):
    root: Annotated[ExplicitBound | FromParamsBound, pd.Field(discriminator="b")]


class TypeDef(ConfiguredBaseModel):
    extension: ExtensionId
    name: str
    description: str
    params: list[TypeParam]
    bound: TypeDefBound

    def deserialize(self, extension: ext.Extension) -> ext.TypeDef:
        return extension.add_type_def(
            ext.TypeDef(
                name=self.name,
                description=self.description,
                params=deser_it(self.params),
                bound=self.bound.root.deserialize(),
            )
        )


class ExtensionValue(ConfiguredBaseModel):
    extension: ExtensionId
    name: str
    typed_value: Value

    def deserialize(self, extension: ext.Extension) -> ext.ExtensionValue:
        return extension.add_extension_value(
            ext.ExtensionValue(
                name=self.name,
                val=self.typed_value.deserialize(),
            )
        )


# --------------------------------------
# --------------- OpDef ----------------
# --------------------------------------


class FixedHugr(ConfiguredBaseModel):
    extensions: ExtensionSet
    hugr: Any

    def deserialize(self) -> ext.FixedHugr:
        return ext.FixedHugr(extensions=self.extensions, hugr=self.hugr)


class OpDef(ConfiguredBaseModel, populate_by_name=True):
    """Serializable definition for dynamically loaded operations."""

    extension: ExtensionId
    name: str  # Unique identifier of the operation.
    description: str  # Human readable description of the operation.
    misc: dict[str, Any] | None = None
    signature: PolyFuncType | None = None
    binary: bool = False
    lower_funcs: list[FixedHugr] = pd.Field(default_factory=list)

    def deserialize(self, extension: ext.Extension) -> ext.OpDef:
        signature = ext.OpDefSig(
            self.signature.deserialize().with_runtime_reqs([extension.name])
            if self.signature
            else None,
            self.binary,
        )

        return extension.add_op_def(
            ext.OpDef(
                name=self.name,
                description=self.description,
                misc=self.misc or {},
                signature=signature,
                lower_funcs=[f.deserialize() for f in self.lower_funcs],
            )
        )


class Extension(ConfiguredBaseModel):
    version: SemanticVersion
    name: ExtensionId
    runtime_reqs: set[ExtensionId]
    types: dict[str, TypeDef]
    values: dict[str, ExtensionValue]
    operations: dict[str, OpDef]

    @classmethod
    def get_version(cls) -> str:
        return serialization_version()

    def deserialize(self) -> ext.Extension:
        e = ext.Extension(
            version=self.version,  # type: ignore[arg-type]
            name=self.name,
            runtime_reqs=self.runtime_reqs,
        )

        for k, t in self.types.items():
            assert k == t.name, "Type name must match key"
            e.add_type_def(t.deserialize(e))

        for k, o in self.operations.items():
            assert k == o.name, "Operation name must match key"
            e.add_op_def(o.deserialize(e))

        for k, v in self.values.items():
            assert k == v.name, "Value name must match key"
            e.add_extension_value(v.deserialize(e))

        return e


class Package(ConfiguredBaseModel):
    modules: list[SerialHugr]
    extensions: list[Extension] = pd.Field(default_factory=list)

    @classmethod
    def get_version(cls) -> str:
        return serialization_version()

    def deserialize(self) -> package.Package:
        return package.Package(
            modules=[Hugr._from_serial(m) for m in self.modules],
            extensions=[e.deserialize() for e in self.extensions],
        )


from hugr import (  # noqa: E402
    ext,
    package,
)



"""Builder classes for HUGR control flow graphs."""

from __future__ import annotations

from contextlib import AbstractContextManager
from dataclasses import dataclass
from typing import TYPE_CHECKING

from typing_extensions import Self

from hugr import ops, tys, val
from hugr.build.base import ParentBuilder
from hugr.build.dfg import DfBase
from hugr.exceptions import MismatchedExit, NoSiblingAncestor, NotInSameCfg
from hugr.hugr import Hugr

if TYPE_CHECKING:
    from hugr.hugr.node_port import Node, PortOffset, ToNode, Wire
    from hugr.tys import Type, TypeRow


class Block(DfBase[ops.DataflowBlock]):
    """Builder class for a basic block in a HUGR control flow graph."""

    def set_outputs(self, *outputs: Wire) -> None:
        super().set_outputs(*outputs)

        assert len(outputs) > 0
        branching = outputs[0]
        branch_type = self.hugr.port_type(branching.out_port())
        assert isinstance(branch_type, tys.Sum)
        self._set_parent_output_count(len(branch_type.variant_rows))

    def set_block_outputs(self, branching: Wire, *other_outputs: Wire) -> None:
        self.set_outputs(branching, *other_outputs)

    def set_single_succ_outputs(self, *outputs: Wire) -> None:
        u = self.load(val.Unit)
        self.set_outputs(u, *outputs)

    def _wire_up_port(self, node: Node, offset: PortOffset, p: Wire) -> Type:
        src = p.out_port()
        cfg_node = self.hugr[self.parent_node].parent
        assert cfg_node is not None
        src_parent = self.hugr[src.node].parent
        try:
            super()._wire_up_port(node, offset, p)
        except NoSiblingAncestor as e:
            # note this just checks if there is a common CFG ancestor
            # it does not check for valid dominance between basic blocks
            # that is deferred to full HUGR validation.
            while cfg_node != src_parent:
                if src_parent is None or src_parent == self.hugr.root:
                    raise NotInSameCfg(src.node.idx, node.idx) from e
                src_parent = self.hugr[src_parent].parent

            self.hugr.add_link(src, node.inp(offset))
        return self._get_dataflow_type(src)


@dataclass
class Cfg(ParentBuilder[ops.CFG], AbstractContextManager):
    """Builder class for a HUGR control flow graph, with the HUGR root node
    being a :class:`CFG <hugr.ops.CFG>`.

    Args:
        input_types: The input types for the CFG. Outputs are computed
        by propagating types through the control flow graph to the exit block.

    Examples:
        >>> cfg = Cfg(tys.Bool, tys.Unit)
        >>> cfg.parent_op
        CFG(inputs=[Bool, Unit])
    """

    #: The HUGR instance this CFG is part of.
    hugr: Hugr
    #: The parent node of the CFG.
    parent_node: Node
    _entry_block: Block
    #: The node holding the root of the exit block.
    exit: Node

    def __init__(self, *input_types: Type) -> None:
        input_typs = list(input_types)
        root_op = ops.CFG(inputs=input_typs)
        hugr = Hugr(root_op)
        self._init_impl(hugr, hugr.root, input_typs)

    def _init_impl(self: Cfg, hugr: Hugr, root: Node, input_types: TypeRow) -> None:
        self.hugr = hugr
        self.parent_node = root
        # to ensure entry is first child, add a dummy entry at the start
        self._entry_block = Block.new_nested(ops.DataflowBlock(input_types), hugr, root)

        self.exit = self.hugr.add_node(ops.ExitBlock(), self.parent_node)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args) -> None:
        return None

    @classmethod
    def new_nested(
        cls,
        input_types: TypeRow,
        hugr: Hugr,
        parent: ToNode | None = None,
    ) -> Cfg:
        """Start building a CFG nested inside an existing HUGR graph.

        Args:
            input_types: The input types for the CFG.
            hugr: The HUGR instance this CFG is part of.
            parent: The parent node for the CFG: defaults to the root of the HUGR
                instance.

        Returns:
            The new CFG builder.

        Examples:
            >>> hugr = Hugr()
            >>> cfg = Cfg.new_nested([tys.Bool], hugr)
            >>> cfg.parent_op
            CFG(inputs=[Bool])
        """
        new = cls.__new__(cls)
        root = hugr.add_node(
            ops.CFG(inputs=input_types),
            parent or hugr.root,
        )
        new._init_impl(hugr, root, input_types)
        return new

    @property
    def entry(self) -> Node:
        """Node for entry block of the CFG.

        Examples:
            >>> cfg = Cfg(tys.Bool)
            >>> cfg.entry
            Node(1)
        """
        return self._entry_block.parent_node

    @property
    def _entry_op(self) -> ops.DataflowBlock:
        return self.hugr._get_typed_op(self.entry, ops.DataflowBlock)

    @property
    def _exit_op(self) -> ops.ExitBlock:
        return self.hugr._get_typed_op(self.exit, ops.ExitBlock)

    def add_entry(self) -> Block:
        """Start building the entry block of the CFG.

        Returns:
            The entry block builder.

        Examples:
            >>> cfg = Cfg(tys.Bool)
            >>> entry = cfg.add_entry()
            >>> entry.set_outputs(*entry.inputs())
        """
        return self._entry_block

    def add_block(self, *input_types: Type) -> Block:
        """Add a new block to the CFG and start building it.

        Args:
            input_types: The input types for the block.

        Returns:
            The block builder.

        Examples:
            >>> cfg = Cfg(tys.Bool)
            >>> with cfg.add_block(tys.Unit) as b:\
                    b.set_single_succ_outputs(*b.inputs())
        """
        new_block = Block.new_nested(
            ops.DataflowBlock(list(input_types)),
            self.hugr,
            self.parent_node,
        )
        return new_block

    # TODO insert_block

    def add_successor(self, pred: Wire) -> Block:
        """Start building a block that succeeds an existing block.

        Args:
            pred: The wire from the predecessor block to the new block. The
            port of the wire determines the branching index of the new block.


        Returns:
            The new block builder.

        Examples:
            >>> cfg = Cfg(tys.Bool)
            >>> with cfg.add_entry() as entry:\
                    entry.set_single_succ_outputs()
            >>> with cfg.add_successor(entry[0]) as b:\
                    b.set_single_succ_outputs(*b.inputs())
        """
        b = self.add_block(*self._nth_outputs(pred))

        self.branch(pred, b)
        return b

    def _nth_outputs(self, wire: Wire) -> TypeRow:
        port = wire.out_port()
        block = self.hugr._get_typed_op(port.node, ops.DataflowBlock)
        return block.nth_outputs(port.offset)

    def branch(self, src: Wire, dst: ToNode) -> None:
        """Add a branching control flow link between blocks.

        Args:
            src: The wire from the predecessor block.
            dst: The destination block.

        Examples:
            >>> cfg = Cfg(tys.Bool)
            >>> with cfg.add_entry() as entry:\
                    entry.set_single_succ_outputs()
            >>> b = cfg.add_block(tys.Unit)
            >>> cfg.branch(entry[0], b)
        """
        # TODO check for existing link/type compatibility
        if dst.to_node() == self.exit:
            return self.branch_exit(src)
        src = src.out_port()
        self.hugr.add_link(src, dst.inp(0))

    def branch_exit(self, src: Wire) -> None:
        """Branch from a block to the exit block.

        Args:
            src: The wire from the predecessor block.

        Examples:
            >>> cfg = Cfg(tys.Bool)
            >>> with cfg.add_entry() as entry:\
                    entry.set_single_succ_outputs()
            >>> cfg.branch_exit(entry[0])
        """
        src = src.out_port()
        self.hugr.add_link(src, self.exit.inp(0))

        out_types = self._nth_outputs(src)
        if self._exit_op._cfg_outputs is not None:
            if self._exit_op._cfg_outputs != out_types:
                raise MismatchedExit(src.node.idx)
        else:
            self._exit_op._cfg_outputs = out_types
            self.parent_op._outputs = out_types
            self.parent_node = self.hugr._update_node_outs(
                self.parent_node, len(out_types)
            )



"""Classes for building HUGRs."""

from .base import ParentBuilder
from .cfg import Block, Cfg
from .cond_loop import Case, Conditional, If, TailLoop
from .dfg import DefinitionBuilder, DfBase, Dfg, Function
from .function import Module
from .tracked_dfg import TrackedDfg

__all__ = [
    "ParentBuilder",
    "Cfg",
    "Block",
    "Case",
    "If",
    "Conditional",
    "TailLoop",
    "Function",
    "Module",
    "TrackedDfg",
    "Dfg",
    "DefinitionBuilder",
    "DfBase",
]



from __future__ import annotations

import json
import os
import pathlib
import subprocess
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, TypeVar

from typing_extensions import Self

from hugr import ext, tys
from hugr._serialization.serial_hugr import SerialHugr
from hugr.envelope import EnvelopeConfig
from hugr.hugr import Hugr
from hugr.ops import AsExtOp, Command, DataflowOp, ExtOp, RegisteredOp
from hugr.std.float import FLOAT_T

if TYPE_CHECKING:
    from syrupy.assertion import SnapshotAssertion

    from hugr.ops import ComWire
    from hugr.package import Package

QUANTUM_EXT = ext.Extension("pytest.quantum,", ext.Version(0, 1, 0))
QUANTUM_EXT.add_op_def(
    ext.OpDef(
        name="H",
        description="Hadamard gate",
        signature=ext.OpDefSig(tys.FunctionType.endo([tys.Qubit])),
    )
)

QUANTUM_EXT.add_op_def(
    ext.OpDef(
        name="CX",
        description="CNOT gate",
        signature=ext.OpDefSig(tys.FunctionType.endo([tys.Qubit] * 2)),
    )
)


E = TypeVar("E", bound=Enum)


def _load_enum(enum_cls: type[E], custom: ExtOp) -> E | None:
    ext = custom._op_def._extension
    assert ext is not None
    if ext.name == QUANTUM_EXT.name and custom._op_def.name in enum_cls.__members__:
        return enum_cls(custom._op_def.name)
    return None


@dataclass(frozen=True)
class OneQbGate(AsExtOp):
    # Have to nest enum to avoid meta class conflict
    class _Enum(Enum):
        H = "H"

    _enum: _Enum

    def __call__(self, q: ComWire) -> Command:
        return DataflowOp.__call__(self, q)

    def op_def(self) -> ext.OpDef:
        return QUANTUM_EXT.operations[self._enum.value]

    @classmethod
    def from_ext(cls, custom: ExtOp) -> Self | None:
        return cls(e) if (e := _load_enum(cls._Enum, custom)) else None


H = OneQbGate(OneQbGate._Enum.H)


@dataclass(frozen=True)
class TwoQbGate(AsExtOp):
    class _Enum(Enum):
        CX = "CX"

    _enum: _Enum

    def op_def(self) -> ext.OpDef:
        return QUANTUM_EXT.operations[self._enum.value]

    @classmethod
    def from_ext(cls, custom: ExtOp) -> Self | None:
        return cls(e) if (e := _load_enum(cls._Enum, custom)) else None

    def __call__(self, q0: ComWire, q1: ComWire) -> Command:
        return DataflowOp.__call__(self, q0, q1)


CX = TwoQbGate(TwoQbGate._Enum.CX)


@QUANTUM_EXT.register_op(
    "Measure",
    signature=tys.FunctionType([tys.Qubit], [tys.Qubit, tys.Bool]),
)
@dataclass(frozen=True)
class MeasureDef(RegisteredOp):
    def __call__(self, q: ComWire) -> Command:
        return super().__call__(q)


Measure = MeasureDef()


@QUANTUM_EXT.register_op(
    "Rz",
    signature=tys.FunctionType([tys.Qubit, FLOAT_T], [tys.Qubit]),
)
@dataclass(frozen=True)
class RzDef(RegisteredOp):
    def __call__(self, q: ComWire, fl_wire: ComWire) -> Command:
        return super().__call__(q, fl_wire)


Rz = RzDef()


def _base_command() -> list[str]:
    workspace_dir = pathlib.Path(__file__).parent.parent.parent
    # use the HUGR_BIN environment variable if set, otherwise use the debug build
    bin_loc = os.environ.get("HUGR_BIN", str(workspace_dir / "target/debug/hugr"))
    return [bin_loc]


def mermaid(h: Hugr):
    """Render the Hugr as a mermaid diagram for debugging."""
    cmd = [*_base_command(), "mermaid", "-"]
    _run_hugr_cmd(h._to_serial().to_json(), cmd)


def validate(
    h: Hugr | Package,
    *,
    roundtrip: bool = True,
    snap: SnapshotAssertion | None = None,
):
    """Validate a HUGR or package.

    args:
        h: The HUGR or package to validate.
        roundtrip: Whether to roundtrip the HUGR through the CLI.
        snapshot: A hugr render snapshot. If not None, it will be compared against the
        rendered HUGR. Pass `--snapshot-update` to pytest to update the snapshot file.
    """
    # TODO: Use envelopes instead of legacy hugr-json
    cmd = [*_base_command(), "validate", "-"]

    if isinstance(h, Hugr):
        cmd += ["--hugr-json"]
        serial = h.to_json()
        _run_hugr_cmd(serial, cmd)
    else:
        serial = h.to_str(EnvelopeConfig.TEXT)
        _run_hugr_cmd(serial, cmd)

    if not roundtrip:
        return
    hugrs = [h] if isinstance(h, Hugr) else h.modules
    for h in hugrs:
        serial = h.to_json()

        starting_json = json.loads(serial)
        h2 = Hugr._from_serial(SerialHugr.load_json(starting_json))
        roundtrip_json = json.loads(h2._to_serial().to_json())
        assert roundtrip_json == starting_json

    if snap is not None and isinstance(h, Hugr):
        dot = h.render_dot()
        assert snap == dot.source
        if os.environ.get("HUGR_RENDER_DOT"):
            dot.pipe("svg")


def _run_hugr_cmd(serial: str, cmd: list[str]):
    try:
        subprocess.run(cmd, check=True, input=serial.encode(), capture_output=True)  # noqa: S603
    except subprocess.CalledProcessError as e:
        error = e.stderr.decode()
        raise RuntimeError(error) from e



from typing import Any

from pydantic import ConfigDict, Field

from hugr.hugr.node_port import NodeIdx, PortOffset

from .ops import OpType
from .ops import classes as ops_classes
from .tys import ConfiguredBaseModel, model_rebuild

Port = tuple[NodeIdx, PortOffset | None]
Edge = tuple[Port, Port]


def serialization_version() -> str:
    """Return the current version of the serialization schema."""
    return "live"


VersionField = Field(
    default_factory=serialization_version,
    title="Version",
    description="Serialisation Schema Version",
    frozen=True,
)


class SerialHugr(ConfiguredBaseModel):
    """A serializable representation of a Hugr."""

    version: str = VersionField
    nodes: list[OpType]
    edges: list[Edge]
    metadata: list[dict[str, Any] | None] | None = None
    encoder: str | None = Field(
        default=None, description="The name of the encoder used to generate the Hugr."
    )

    def to_json(self) -> str:
        """Return a JSON representation of the Hugr."""
        from hugr import __version__ as hugr_version

        self.encoder = f"hugr-py v{hugr_version}"
        return self.model_dump_json()

    @classmethod
    def load_json(cls, json: dict[Any, Any]) -> "SerialHugr":
        """Decode a JSON-encoded Hugr."""
        return cls(**json)

    @classmethod
    def get_version(cls) -> str:
        """Return the version of the schema."""
        return cls(nodes=[], edges=[]).version

    @classmethod
    def _pydantic_rebuild(cls, config: ConfigDict | None = None, **kwargs):
        config = config or ConfigDict()
        my_classes = dict(ops_classes)
        my_classes[cls.__name__] = cls
        model_rebuild(my_classes, config=config, **kwargs)

    model_config = ConfigDict(
        title="Hugr",
        json_schema_extra={
            "required": ["version", "nodes", "edges"],
        },
    )



from __future__ import annotations

import pytest

from hugr import tys
from hugr.val import FALSE, TRUE, Left, None_, Right, Some, Sum, Tuple, UnitSum, Value


def test_sums():
    assert Sum(0, tys.Tuple(), []) == Tuple()
    assert Sum(0, tys.Tuple(tys.Bool, tys.Bool), [TRUE, FALSE]) == Tuple(TRUE, FALSE)

    ty = tys.Sum([[], [tys.Bool, tys.Bool]])
    assert Sum(1, ty, [TRUE, FALSE]) == Some(TRUE, FALSE)
    assert Sum(1, ty, [TRUE, FALSE]) == Right([], [TRUE, FALSE])
    assert Sum(0, ty, []) == None_(tys.Bool, tys.Bool)
    assert Sum(0, ty, []) == Left([], [tys.Bool, tys.Bool])

    ty = tys.Sum([[tys.Bool], [tys.Bool]])
    assert Sum(0, ty, [TRUE]) == Left([TRUE], [tys.Bool])
    assert Sum(1, ty, [FALSE]) == Right([tys.Bool], [FALSE])

    assert Tuple() == Sum(0, tys.Tuple(), [])
    assert UnitSum(0, size=1) == Tuple()
    assert UnitSum(2, size=4) == Sum(2, tys.UnitSum(size=4), [])


@pytest.mark.parametrize(
    ("value", "string", "repr_str"),
    [
        (
            Sum(0, tys.Sum([[tys.Bool], [tys.Qubit]]), [TRUE, FALSE]),
            "Sum(tag=0, typ=Sum([[Bool], [Qubit]]), vals=[TRUE, FALSE])",
            "Sum(tag=0, typ=Sum([[Bool], [Qubit]]), vals=[TRUE, FALSE])",
        ),
        (UnitSum(0, size=1), "Unit", "Unit"),
        (UnitSum(0, size=2), "FALSE", "FALSE"),
        (UnitSum(1, size=2), "TRUE", "TRUE"),
        (UnitSum(2, size=5), "UnitSum(2, 5)", "UnitSum(2, 5)"),
        (Tuple(TRUE, FALSE), "Tuple(TRUE, FALSE)", "Tuple(TRUE, FALSE)"),
        (Some(TRUE, FALSE), "Some(TRUE, FALSE)", "Some(TRUE, FALSE)"),
        (None_(tys.Bool, tys.Bool), "None", "None(Bool, Bool)"),
        (
            Left([TRUE, FALSE], [tys.Bool]),
            "Left(TRUE, FALSE)",
            "Left(vals=[TRUE, FALSE], right_typ=[Bool])",
        ),
        (
            Right([tys.Bool, tys.Bool], [FALSE]),
            "Right(FALSE)",
            "Right(left_typ=[Bool, Bool], vals=[FALSE])",
        ),
    ],
)
def test_val_sum_str(value: Value, string: str, repr_str: str):
    assert str(value) == string
    assert repr(value) == repr_str



from pydantic_extra_types.semantic_version import SemanticVersion

from hugr._serialization.extension import (
    ExplicitBound,
    Extension,
    OpDef,
    Package,
    TypeDef,
    TypeDefBound,
)
from hugr._serialization.ops import Module, OpType
from hugr._serialization.serial_hugr import SerialHugr, serialization_version
from hugr._serialization.tys import (
    FunctionType,
    PolyFuncType,
    Type,
    TypeBound,
    TypeParam,
    TypeTypeParam,
    Variable,
)
from hugr.envelope import EnvelopeConfig

EXAMPLE = r"""
{
    "version": "0.1.0",
    "name": "ext",
    "runtime_reqs": [],
    "types": {
        "foo": {
            "extension": "ext",
            "name": "foo",
            "params": [
                {
                    "tp": "Type",
                    "b": "C"
                }
            ],
            "description": "foo",
            "bound": {
                "b": "Explicit",
                "bound": "C"
            }
        }
    },
    "values": {},
    "operations": {
        "New": {
            "extension": "ext",
            "name": "New",
            "description": "new",
            "signature": {
                "params": [
                    {
                        "tp": "Type",
                        "b": "C"
                    }
                ],
                "body": {
                    "input": [
                        {
                            "t": "V",
                            "i": 0,
                            "b": "C"
                        }
                    ],
                    "output": [],
                    "runtime_reqs": []
                }
            },
            "lower_funcs": []
        }
    }
}
"""


def test_extension():
    assert serialization_version() == Extension.get_version()
    param = TypeParam(root=TypeTypeParam(b=TypeBound.Copyable))

    bound = TypeDefBound(root=ExplicitBound(bound=TypeBound.Copyable))
    type_def = TypeDef(
        extension="ext", name="foo", description="foo", params=[param], bound=bound
    )
    body = FunctionType(
        input=[Type(root=Variable(b=TypeBound.Copyable, i=0))], output=[]
    )
    op_def = OpDef(
        extension="ext",
        name="New",
        description="new",
        signature=PolyFuncType(
            params=[param],
            body=body,
        ),
        lower_funcs=[],
    )
    ext = Extension(
        version=SemanticVersion(0, 1, 0),
        name="ext",
        runtime_reqs=set(),
        types={"foo": type_def},
        values={},
        operations={"New": op_def},
    )

    ext_load = Extension.model_validate_json(EXAMPLE)
    assert ext == ext_load

    dumped_json = ext.model_dump_json()

    assert Extension.model_validate_json(dumped_json) == ext
    hugr_ext = ext.deserialize()
    assert hugr_ext.from_json(hugr_ext.to_json()) == hugr_ext


def test_package():
    assert serialization_version() == Package.get_version()

    ext = Extension(
        version=SemanticVersion(0, 1, 0),
        name="ext",
        runtime_reqs=set(),
        types={},
        values={},
        operations={},
    )
    ext_load = Extension.model_validate_json(EXAMPLE)

    package = Package(
        extensions=[ext, ext_load],
        modules=[SerialHugr(nodes=[OpType(root=Module(parent=0))], edges=[])],
    )

    package_load = Package.model_validate_json(package.model_dump_json())
    assert package == package_load

    hugr_package = package.deserialize()
    assert (
        hugr_package.from_bytes(hugr_package.to_bytes(EnvelopeConfig.TEXT))
        == hugr_package
    )



import pytest

from hugr.utils import BiMap, NotBijection


def test_insert_left() -> None:
    bimap: BiMap[str, int] = BiMap()
    bimap.insert_left("a", 1)
    assert bimap["a"] == 1
    assert bimap.get_left(1) == "a"


def test_insert_right() -> None:
    bimap: BiMap[str, int] = BiMap()
    bimap.insert_right(1, "a")
    assert bimap["a"] == 1
    assert bimap.get_left(1) == "a"


def test_delete_left() -> None:
    bimap: BiMap[str, int] = BiMap()
    bimap.insert_left("a", 1)
    del bimap["a"]
    assert bimap.get_right("a") is None
    assert bimap.get_left(1) is None


def test_delete_right() -> None:
    bimap: BiMap[str, int] = BiMap()
    bimap.insert_right(1, "a")
    bimap.delete_right(1)
    assert bimap.get_right("a") is None
    assert bimap.get_left(1) is None


def test_iter() -> None:
    bimap: BiMap[str, int] = BiMap()
    bimap.insert_left("a", 1)
    bimap.insert_left("b", 2)
    bimap.insert_left("c", 3)
    assert set(bimap) == {"a", "b", "c"}
    assert list(bimap.items()) == [("a", 1), ("b", 2), ("c", 3)]


def test_len() -> None:
    bimap: BiMap[str, int] = BiMap()
    assert len(bimap) == 0
    bimap.insert_left("a", 1)
    assert len(bimap) == 1
    bimap.insert_left("b", 2)
    assert len(bimap) == 2

    bimap.delete_left("a")
    assert len(bimap) == 1


def test_existing_key() -> None:
    bimap: BiMap[str, int] = BiMap()
    bimap.insert_left("a", 1)
    bimap.insert_left("b", 1)

    assert bimap.get_right("b") == 1
    assert bimap.get_left(1) == "b"

    assert bimap.get_right("a") is None


def test_bimap_init():
    # Test with empty initial map
    bm = BiMap()
    assert len(bm) == 0

    # Test with non-empty initial map
    initial_map = {"a": 1, "b": 2}
    bm = BiMap(initial_map)
    assert len(bm) == 2

    # Test with non-bijection initial map
    invalid_map = {"a": 1, "b": 1}
    with pytest.raises(NotBijection):
        bm = BiMap(invalid_map)



from hugr.std.prelude import STRING_T, StringVal


def test_string_val():
    ext_val = StringVal("test").to_value()
    assert ext_val.name == "ConstString"
    assert ext_val.typ == STRING_T
    assert ext_val.val == {"value": "test"}



"""Visualise HUGR using graphviz."""

from collections.abc import Iterable
from dataclasses import dataclass, field

import graphviz as gv  # type: ignore[import-untyped]
from graphviz import Digraph
from typing_extensions import assert_never

from hugr.hugr import Hugr
from hugr.ops import AsExtOp
from hugr.tys import CFKind, ConstKind, FunctionKind, Kind, OrderKind, ValueKind

from .node_port import InPort, Node, OutPort


@dataclass(frozen=True)
class Palette:
    """A set of colours used for rendering."""

    background: str
    node: str
    edge: str
    dark: str
    const: str
    discard: str
    node_border: str
    port_border: str

    @classmethod
    def named(cls, name: str) -> "Palette":
        return PALETTE[name]


PALETTE: dict[str, Palette] = {
    "default": Palette(
        background="white",
        node="#ACCBF9",
        edge="#1CADE4",
        dark="black",
        const="#77CEEF",
        discard="#ff8888",
        node_border="white",
        port_border="#1CADE4",
    ),
    "nb": Palette(
        background="white",
        node="#7952B3",
        edge="#FFC107",
        dark="#343A40",
        const="#7c55b4",
        discard="#ff8888",
        node_border="#9d80c7",
        port_border="#ffd966",
    ),
    "zx": Palette(
        background="white",
        node="#629DD1",
        edge="#297FD5",
        dark="#112D4E",
        const="#a1eea1",
        discard="#ff8888",
        node_border="#D8F8D8",
        port_border="#E8A5A5",
    ),
}


@dataclass
class RenderConfig:
    """Configuration for rendering a HUGR to a graphviz dot file."""

    #: The palette to use for rendering. See :obj:`PALETTE` for the included options.
    palette: Palette = field(default_factory=lambda: PALETTE["default"])
    #: If true prepend extension name to operation name.
    qualify_op_name: bool = False


class DotRenderer:
    """Render a HUGR to a graphviz dot file.

    Args:
        config: Render config
    """

    config: RenderConfig

    def __init__(self, config: RenderConfig | None = None) -> None:
        self.config = config or RenderConfig()

    def render(self, hugr: Hugr) -> Digraph:
        """Render a HUGR to a graphviz dot object."""
        graph_attr = {
            "rankdir": "",
            "ranksep": "0.1",
            "nodesep": "0.15",
            "margin": "0",
            "bgcolor": self.config.palette.background,
        }
        if not (name := hugr[hugr.root].metadata.get("name", None)):
            name = ""

        graph = gv.Digraph(name, strict=False)
        graph.attr(**graph_attr)

        self._viz_node(hugr.root, hugr, graph)

        for src_port, tgt_port in hugr.links():
            kind = hugr.port_kind(src_port)
            self._viz_link(src_port, tgt_port, kind, graph)

        return graph

    def store(self, hugr: Hugr, filename: str, format: str = "svg") -> None:
        """Render a HUGR and save it to a file.

        Args:
            hugr: The HUGR to render.
            filename: Filename for saving the rendered graph.
            format: The format used for rendering ('pdf', 'png', etc.).
                Defaults to SVG.
        """
        gv_graph = self.render(hugr)
        gv_graph.render(filename, format=format)

    _FONTFACE = "monospace"

    _HTML_LABEL_TEMPLATE = """
    <TABLE BORDER="{border_width}" CELLBORDER="0" CELLSPACING="1" CELLPADDING="1"
        BGCOLOR="{node_back_color}" COLOR="{border_colour}">
    {inputs_row}
    <TR>
        <TD>
        <TABLE BORDER="0" CELLBORDER="0">
            <TR><TD><FONT POINT-SIZE="{fontsize}" FACE="{fontface}"
                COLOR="{label_color}"><B>{node_label}</B>{node_data}</FONT></TD></TR>
        </TABLE>
        </TD>
    </TR>
    {outputs_row}
    </TABLE>
    """

    _HTML_PORTS_ROW_TEMPLATE = """
        <TR>
            <TD>
                <TABLE BORDER="0" CELLBORDER="0" CELLSPACING="3" CELLPADDING="2">
                    <TR>
                        {port_cells}
                    </TR>
                </TABLE>
            </TD>
        </TR>
    """

    _HTML_PORT_TEMPLATE = (
        '<TD BGCOLOR="{back_colour}" COLOR="{border_colour}"'
        ' PORT="{port_id}" BORDER="{border_width}">'
        '<FONT POINT-SIZE="10.0" FACE="{fontface}" COLOR="{font_colour}">{port}</FONT>'
        "</TD>"
    )

    _INPUT_PREFIX = "in."
    _OUTPUT_PREFIX = "out."

    def _format_html_label(self, **kwargs: str) -> str:
        _HTML_LABEL_DEFAULTS = {
            "label_color": self.config.palette.dark,
            "node_back_color": self.config.palette.node,
            "inputs_row": "",
            "outputs_row": "",
            "border_colour": self.config.palette.port_border,
            "border_width": "1",
            "fontface": self._FONTFACE,
            "fontsize": 11.0,
        }
        return self._HTML_LABEL_TEMPLATE.format(**{**_HTML_LABEL_DEFAULTS, **kwargs})

    def _html_ports(self, ports: Iterable[str], id_prefix: str) -> str:
        return self._HTML_PORTS_ROW_TEMPLATE.format(
            port_cells="".join(
                self._HTML_PORT_TEMPLATE.format(
                    port=port,
                    # differentiate input and output node identifiers
                    # with a prefix
                    port_id=id_prefix + port,
                    back_colour=self.config.palette.background,
                    font_colour=self.config.palette.dark,
                    border_width="1",
                    border_colour=self.config.palette.port_border,
                    fontface=self._FONTFACE,
                )
                for port in ports
            )
        )

    def _in_port_name(self, p: InPort) -> str:
        return f"{p.node.idx}:{self._INPUT_PREFIX}{p.offset}"

    def _out_port_name(self, p: OutPort) -> str:
        return f"{p.node.idx}:{self._OUTPUT_PREFIX}{p.offset}"

    def _in_order_name(self, n: Node) -> str:
        return f"{n.idx}:{self._INPUT_PREFIX}None"

    def _out_order_name(self, n: Node) -> str:
        return f"{n.idx}:{self._OUTPUT_PREFIX}None"

    def _viz_node(self, node: Node, hugr: Hugr, graph: Digraph) -> None:
        """Render a (possibly nested) node to a graphviz graph."""
        meta = hugr[node].metadata
        if len(meta) > 0:
            data = "<BR/><BR/>" + "<BR/>".join(
                f"{key}: {value}" for key, value in meta.items()
            )
        else:
            data = ""

        in_ports = [str(i) for i in range(hugr.num_in_ports(node))]
        out_ports = [str(i) for i in range(hugr.num_out_ports(node))]
        inputs_row = (
            self._html_ports(in_ports, self._INPUT_PREFIX) if len(in_ports) > 0 else ""
        )
        outputs_row = (
            self._html_ports(out_ports, self._OUTPUT_PREFIX)
            if len(out_ports) > 0
            else ""
        )

        op = hugr[node].op
        if isinstance(op, AsExtOp) and not self.config.qualify_op_name:
            op_name = op.op_def().name
        else:
            op_name = op.name()
        if hugr.children(node):
            with graph.subgraph(name=f"cluster{node.idx}") as sub:
                for child in hugr.children(node):
                    self._viz_node(child, hugr, sub)
                html_label = self._format_html_label(
                    node_back_color=self.config.palette.edge,
                    node_label=op_name,
                    node_data=data,
                    border_colour=self.config.palette.port_border,
                    inputs_row=inputs_row,
                    outputs_row=outputs_row,
                )
                sub.node(f"{node.idx}", shape="plain", label=f"<{html_label}>")
                sub.attr(label="", margin="10", color=self.config.palette.edge)
        else:
            html_label = self._format_html_label(
                node_back_color=self.config.palette.node,
                node_label=op_name,
                node_data=data,
                inputs_row=inputs_row,
                outputs_row=outputs_row,
                border_colour=self.config.palette.background,
            )
            graph.node(f"{node.idx}", label=f"<{html_label}>", shape="plain")

    def _viz_link(
        self, src_port: OutPort, tgt_port: InPort, kind: Kind, graph: Digraph
    ) -> None:
        edge_attr = {
            "penwidth": "1.5",
            "arrowhead": "none",
            "arrowsize": "1.0",
            "fontname": self._FONTFACE,
            "fontsize": "9",
            "fontcolor": "black",
        }

        label = ""
        match kind:
            case ValueKind(ty):
                label = str(ty)
                color = self.config.palette.edge
            case OrderKind():
                color = self.config.palette.dark
            case ConstKind() | FunctionKind():
                color = self.config.palette.const
            case CFKind():
                color = self.config.palette.dark
            case _:
                assert_never(kind)

        graph.edge(
            self._out_port_name(src_port),
            self._in_port_name(tgt_port),
            label=label,
            color=color,
            **edge_attr,
        )



from hugr import Hugr, tys
from hugr._serialization.serial_hugr import SerialHugr, serialization_version
from hugr.build.function import Module


def test_empty():
    h = SerialHugr(nodes=[], edges=[])
    assert h.model_dump() == {
        "version": serialization_version(),
        "nodes": [],
        "edges": [],
        "metadata": None,
        "encoder": None,
    }


def test_children():
    mod = Module()
    mod.declare_function("foo", tys.PolyFuncType([], tys.FunctionType.empty()))

    h = mod.hugr
    assert len(h.children(h.root)) == 1

    h2 = Hugr.load_json(h.to_json())

    assert len(h2.children(h2.root)) == 1



"""Node and port classes for Hugr graphs."""

from __future__ import annotations

from dataclasses import dataclass, field, replace
from enum import Enum
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Generic,
    Protocol,
    TypeVar,
    overload,
)

from typing_extensions import Self

if TYPE_CHECKING:
    from collections.abc import Iterator


class Direction(Enum):
    """Enum over port directions, INCOMING and OUTGOING."""

    INCOMING = 0
    OUTGOING = 1


NodeIdx = int
PortOffset = int


@dataclass(frozen=True, eq=True, order=True)
class _Port:
    node: Node
    offset: PortOffset
    direction: ClassVar[Direction]


@dataclass(frozen=True, eq=True, order=True)
class InPort(_Port):
    """Incoming port, defined by the `node` it belongs to and the port `offset`."""

    direction: ClassVar[Direction] = Direction.INCOMING

    def __repr__(self) -> str:
        return f"InPort({self.node}, {self.offset})"


class Wire(Protocol):
    """Protocol for objects that can provide a dataflow output port."""

    def out_port(self) -> OutPort:
        """OutPort corresponding to this :class:`Wire`."""
        ...  # pragma: no cover


@dataclass(frozen=True, eq=True, order=True)
class OutPort(_Port, Wire):
    """Outgoing port, defined by the `node` it belongs to and the port `offset`."""

    direction: ClassVar[Direction] = Direction.OUTGOING

    def out_port(self) -> OutPort:
        return self

    def __repr__(self) -> str:
        return f"OutPort({self.node}, {self.offset})"


class ToNode(Wire, Protocol):
    """Protocol by any object that can be treated as a :class:`Node`."""

    def to_node(self) -> Node:
        """Convert to a :class:`Node`."""
        ...  # pragma: no cover

    @overload
    def __getitem__(self, index: PortOffset) -> OutPort: ...
    @overload
    def __getitem__(self, index: slice) -> Iterator[OutPort]: ...
    @overload
    def __getitem__(self, index: tuple[PortOffset, ...]) -> Iterator[OutPort]: ...

    def __getitem__(
        self, index: PortOffset | slice | tuple[PortOffset, ...]
    ) -> OutPort | Iterator[OutPort]:
        return self.to_node()._index(index)

    def out_port(self) -> OutPort:
        return OutPort(self.to_node(), 0)

    def outputs(self) -> Iterator[OutPort]:
        """Returns an iterator over the output ports of this node."""
        return self[:]

    def __iter__(self) -> Iterator[OutPort]:
        return self.outputs()

    def inp(self, offset: PortOffset) -> InPort:
        """Generate an input port for this node.

        Args:
            offset: port offset.

        Returns:
            Incoming port for this node.

        Examples:
            >>> Node(0).inp(1)
            InPort(Node(0), 1)
        """
        return InPort(self.to_node(), offset)

    def out(self, offset: PortOffset) -> OutPort:
        """Generate an output port for this node.

        Args:
            offset: port offset.

        Returns:
            Outgoing port for this node.

        Examples:
            >>> Node(0).out(1)
            OutPort(Node(0), 1)
        """
        return OutPort(self.to_node(), offset)

    def port(self, offset: PortOffset, direction: Direction) -> InPort | OutPort:
        """Generate a port in `direction` for this node with `offset`.

        Examples:
            >>> Node(0).port(1, Direction.INCOMING)
            InPort(Node(0), 1)
            >>> Node(0).port(1, Direction.OUTGOING)
            OutPort(Node(0), 1)
        """
        if direction == Direction.INCOMING:
            return self.inp(offset)
        else:
            return self.out(offset)

    @property
    def metadata(self) -> dict[str, object]:
        """Metadata associated with this node."""
        return self.to_node()._metadata


@dataclass(frozen=True, eq=True, order=True)
class Node(ToNode):
    """Node in hierarchical :class:`Hugr <hugr.hugr.Hugr>` graph,
    with globally unique index.
    """

    idx: NodeIdx
    _metadata: dict[str, object] = field(
        repr=False, compare=False, default_factory=dict
    )
    _num_out_ports: int | None = field(default=None, compare=False, repr=False)

    def _index(
        self, index: PortOffset | slice | tuple[PortOffset, ...]
    ) -> OutPort | Iterator[OutPort]:
        match index:
            case PortOffset(index):
                index = self._normalize_index(index)
                return self.out(index)
            case slice():
                start = index.start or 0
                stop = index.stop if index.stop is not None else self._num_out_ports
                if stop is None:
                    msg = (
                        f"{self} does not have a fixed number of output ports. "
                        "Iterating over all output ports is not supported."
                    )
                    raise ValueError(msg)

                start = self._normalize_index(start, allow_overflow=True)
                stop = self._normalize_index(stop, allow_overflow=True)
                step = index.step or 1

                return (self[i] for i in range(start, stop, step))
            case tuple(xs):
                return (self[i] for i in xs)

    def _normalize_index(self, index: int, allow_overflow: bool = False) -> int:
        """Given an index passed to `__getitem__`, normalize it to be within the
        range of output ports.

        Args:
            index: index to normalize.
            allow_overflow: whether to allow indices beyond the number of outputs.
                If True, indices over `self._num_out_ports` will be truncated.

        Returns:
            Normalized index.

        Raises:
            IndexError: if the index is out of range.
        """
        msg = f"Index {index} out of range"

        if self._num_out_ports is not None:
            if index >= self._num_out_ports and not allow_overflow:
                raise IndexError(msg)
            if index < -self._num_out_ports:
                raise IndexError(msg)
        else:
            if index < 0:
                raise IndexError(msg)

        if index >= 0 and self._num_out_ports is not None:
            return min(index, self._num_out_ports)
        elif index >= 0:
            return index
        else:
            assert self._num_out_ports is not None
            return self._num_out_ports + index

    def to_node(self) -> Node:
        return self

    def __repr__(self) -> str:
        return f"Node({self.idx})"


P = TypeVar("P", InPort, OutPort)


@dataclass(frozen=True, eq=True, order=True)
class _SubPort(Generic[P]):
    port: P
    sub_offset: int = 0

    def next_sub_offset(self) -> Self:
        return replace(self, sub_offset=self.sub_offset + 1)



from __future__ import annotations

import pytest

from hugr import val
from hugr.std.collections.array import Array, ArrayVal
from hugr.std.collections.list import List, ListVal
from hugr.std.float import FLOAT_T
from hugr.std.int import INT_T, _int_tv
from hugr.tys import (
    Alias,
    Bool,
    BoundedNatArg,
    BoundedNatParam,
    Either,
    ExtensionsArg,
    ExtensionsParam,
    ExtType,
    FunctionType,
    ListParam,
    Option,
    PolyFuncType,
    Qubit,
    RowVariable,
    SequenceArg,
    StringArg,
    StringParam,
    Sum,
    Tuple,
    TupleParam,
    Type,
    TypeArg,
    TypeBound,
    TypeParam,
    TypeTypeArg,
    TypeTypeParam,
    UnitSum,
    USize,
    Variable,
    VariableArg,
)


def test_sums():
    assert Sum([[Bool, Qubit]]) == Tuple(Bool, Qubit)
    assert Tuple(Bool, Qubit) == Sum([[Bool, Qubit]])
    assert Sum([[Bool, Qubit]]).as_tuple() == Sum([[Bool, Qubit]])

    assert Sum([[], [Bool, Qubit]]) == Option(Bool, Qubit)
    assert Sum([[], [Bool, Qubit]]) == Either([], [Bool, Qubit])
    assert Option(Bool, Qubit) == Either([], [Bool, Qubit])
    assert Sum([[Qubit], [Bool]]) == Either([Qubit], [Bool])

    assert Tuple() == Sum([[]])
    assert UnitSum(0) == Sum([])
    assert UnitSum(1) == Tuple()
    assert UnitSum(4) == Sum([[], [], [], []])


@pytest.mark.parametrize(
    ("ty", "string", "repr_str"),
    [
        (
            Sum([[Bool], [Qubit], [Qubit, Bool]]),
            "Sum([[Bool], [Qubit], [Qubit, Bool]])",
            "Sum([[Bool], [Qubit], [Qubit, Bool]])",
        ),
        (UnitSum(1), "Unit", "Unit"),
        (UnitSum(2), "Bool", "Bool"),
        (UnitSum(3), "UnitSum(3)", "UnitSum(3)"),
        (Tuple(Bool, Qubit), "Tuple(Bool, Qubit)", "Tuple(Bool, Qubit)"),
        (Option(Bool, Qubit), "Option(Bool, Qubit)", "Option(Bool, Qubit)"),
        (
            Either([Bool, Qubit], [Bool]),
            "Either((Bool, Qubit), Bool)",
            "Either(left=[Bool, Qubit], right=[Bool])",
        ),
    ],
)
def test_tys_sum_str(ty: Type, string: str, repr_str: str):
    assert str(ty) == string
    assert repr(ty) == repr_str


@pytest.mark.parametrize(
    ("param", "string"),
    [
        (TypeTypeParam(TypeBound.Any), "Any"),
        (BoundedNatParam(3), "Nat(3)"),
        (BoundedNatParam(None), "Nat"),
        (StringParam(), "String"),
        (
            TupleParam([TypeTypeParam(TypeBound.Any), BoundedNatParam(3)]),
            "(Any, Nat(3))",
        ),
        (ListParam(StringParam()), "[String]"),
        (ExtensionsParam(), "Extensions"),
    ],
)
def test_params_str(param: TypeParam, string: str):
    assert str(param) == string


@pytest.mark.parametrize(
    ("arg", "string"),
    [
        (TypeTypeArg(Bool), "Type(Bool)"),
        (BoundedNatArg(3), "3"),
        (StringArg("hello"), '"hello"'),
        (
            SequenceArg([TypeTypeArg(Qubit), BoundedNatArg(3)]),
            "(Type(Qubit), 3)",
        ),
        (VariableArg(2, StringParam()), "$2"),
        (ExtensionsArg(["A", "B"]), "Extensions(A, B)"),
    ],
)
def test_args_str(arg: TypeArg, string: str):
    assert str(arg) == string


@pytest.mark.parametrize(
    ("ty", "string"),
    [
        (Array(Bool, 3), "array<3, Type(Bool)>"),
        (Variable(2, TypeBound.Any), "$2"),
        (RowVariable(4, TypeBound.Copyable), "$4"),
        (USize(), "USize"),
        (INT_T, "int<5>"),
        (FLOAT_T, "float64"),
        (Alias("Foo", TypeBound.Copyable), "Foo"),
        (FunctionType([Bool, Qubit], [Qubit, Bool]), "Bool, Qubit -> Qubit, Bool"),
        (
            PolyFuncType(
                [TypeTypeParam(TypeBound.Any), BoundedNatParam(7)],
                FunctionType([_int_tv(1)], [Variable(0, TypeBound.Copyable)]),
            ),
            "∀ Any, Nat(7). int<$1> -> $0",
        ),
    ],
)
def test_tys_str(ty: Type, string: str):
    assert str(ty) == string
    if isinstance(ty, ExtType):
        assert str(ty._to_opaque()) == string


def test_list():
    ty_var = Variable(0, TypeBound.Copyable)

    ls = List(Bool)
    assert ls.ty == Bool

    ls = List(ty_var)
    assert ls.ty == ty_var

    l_val = ListVal([val.TRUE, val.FALSE], Bool)
    assert l_val.v == [val.TRUE, val.FALSE]
    assert l_val.ty == List(Bool)


def test_array():
    ty_var = Variable(0, TypeBound.Copyable)
    len_var = VariableArg(1, BoundedNatParam())

    ls = Array(Bool, 3)
    assert ls.ty == Bool
    assert ls.size == 3

    ls = Array(ty_var, len_var)
    assert ls.ty == ty_var
    assert ls.size is None

    ar_val = ArrayVal([val.TRUE, val.FALSE], Bool)
    assert ar_val.v == [val.TRUE, val.FALSE]
    assert ar_val.ty == Array(Bool, 2)



from dataclasses import dataclass

import pytest

from hugr import ext, ops, tys
from hugr.build.dfg import Dfg
from hugr.hugr import Hugr, Node
from hugr.ops import AsExtOp, Custom, ExtOp
from hugr.package import Package
from hugr.std.float import FLOAT_T
from hugr.std.float import FLOAT_TYPES_EXTENSION as FLOAT_EXT
from hugr.std.int import INT_OPS_EXTENSION, INT_TYPES_EXTENSION, DivMod, int_t
from hugr.std.logic import EXTENSION as LOGIC_EXT
from hugr.std.logic import Not

from .conftest import CX, QUANTUM_EXT, H, Measure, Rz, validate

STRINGLY_EXT = ext.Extension("my_extension", ext.Version(0, 0, 0))
_STRINGLY_DEF = STRINGLY_EXT.add_op_def(
    ext.OpDef(
        "StringlyOp",
        signature=ext.OpDefSig(
            tys.PolyFuncType([tys.StringParam()], tys.FunctionType.endo([]))
        ),
    )
)


@dataclass
class StringlyOp(AsExtOp):
    tag: str

    def op_def(self) -> ext.OpDef:
        return STRINGLY_EXT.get_op("StringlyOp")

    def type_args(self) -> list[tys.TypeArg]:
        return [tys.StringArg(self.tag)]

    def cached_signature(self) -> tys.FunctionType | None:
        return tys.FunctionType.endo([], runtime_reqs=[STRINGLY_EXT.name])

    @classmethod
    def from_ext(cls, custom: ops.ExtOp) -> "StringlyOp":
        match custom:
            case ops.ExtOp(
                _op_def=_STRINGLY_DEF,
                args=[tys.StringArg(tag)],
            ):
                return cls(tag=tag)
            case _:
                msg = f"Invalid custom op: {custom}"
                raise AsExtOp.InvalidExtOp(msg)


def test_stringly_typed():
    dfg = Dfg()
    n = dfg.add(StringlyOp("world")())
    dfg.set_outputs()
    assert dfg.hugr[n].op == StringlyOp("world")
    validate(Package([dfg.hugr], [STRINGLY_EXT]))

    new_h = Hugr._from_serial(dfg.hugr._to_serial())

    assert isinstance(new_h[n].op, Custom)

    registry = ext.ExtensionRegistry()
    new_h.resolve_extensions(registry)

    # doesn't resolve without extension
    assert isinstance(new_h[n].op, Custom)

    registry.add_extension(STRINGLY_EXT)
    new_h.resolve_extensions(registry)

    assert isinstance(new_h[n].op, ExtOp)


def test_registry():
    reg = ext.ExtensionRegistry()
    reg.add_extension(LOGIC_EXT)
    assert reg.get_extension(LOGIC_EXT.name).name == LOGIC_EXT.name
    assert len(reg.extensions) == 1
    with pytest.raises(ext.ExtensionRegistry.ExtensionExists):
        reg.add_extension(LOGIC_EXT)

    with pytest.raises(ext.ExtensionRegistry.ExtensionNotFound):
        reg.get_extension("not_found")


@pytest.fixture
def registry() -> ext.ExtensionRegistry:
    reg = ext.ExtensionRegistry()
    reg.add_extension(LOGIC_EXT)
    reg.add_extension(QUANTUM_EXT)
    reg.add_extension(STRINGLY_EXT)
    reg.add_extension(INT_TYPES_EXTENSION)
    reg.add_extension(INT_OPS_EXTENSION)
    reg.add_extension(FLOAT_EXT)

    return reg


@pytest.mark.parametrize(
    "as_ext",
    [Not, DivMod, H, CX, Measure, Rz, StringlyOp("hello")],
)
def test_custom_op(as_ext: AsExtOp, registry: ext.ExtensionRegistry):
    ext_op = as_ext.ext_op

    assert ExtOp.from_ext(ext_op) == ext_op

    assert type(as_ext).from_ext(ext_op) == as_ext
    custom = as_ext._to_serial(Node(0)).deserialize()
    assert isinstance(custom, Custom)
    # ExtOp compared to Custom via `to_custom`
    assert custom.resolve(registry) == ext_op
    assert ext_op == as_ext
    assert as_ext == ext_op


def test_custom_bad_eq():
    assert Not != DivMod

    bad_custom_sig = Custom("Not", extension=LOGIC_EXT.name)  # empty signature

    assert Not != bad_custom_sig

    bad_custom_args = Custom(
        "Not",
        extension=LOGIC_EXT.name,
        signature=tys.FunctionType.endo([tys.Bool]),
        args=[tys.Bool.type_arg()],
    )

    assert Not != bad_custom_args


_LIST_T = STRINGLY_EXT.add_type_def(
    ext.TypeDef(
        "List",
        description="A list of elements.",
        params=[tys.TypeTypeParam(tys.TypeBound.Any)],
        bound=ext.FromParamsBound([0]),
    )
)

_BOOL_LIST_T = _LIST_T.instantiate([tys.Bool.type_arg()])


@pytest.mark.parametrize(
    "ext_t",
    [FLOAT_T, int_t(5), _BOOL_LIST_T],
)
def test_custom_type(ext_t: tys.ExtType, registry: ext.ExtensionRegistry):
    opaque = ext_t._to_serial().deserialize()
    assert isinstance(opaque, tys.Opaque)
    assert opaque.resolve(registry) == ext_t

    assert opaque.resolve(ext.ExtensionRegistry()) == opaque

    f_t = tys.FunctionType.endo([ext_t])
    f_t_opaque = f_t._to_serial().deserialize()
    assert isinstance(f_t_opaque.input[0], tys.Opaque)

    assert f_t_opaque.resolve(registry) == f_t



from __future__ import annotations

import pytest

import hugr.ops as ops
import hugr.tys as tys
import hugr.val as val
from hugr.build.dfg import Dfg, _ancestral_sibling
from hugr.build.function import Module
from hugr.hugr import Hugr
from hugr.hugr.node_port import Node, _SubPort
from hugr.ops import NoConcreteFunc
from hugr.std.int import INT_T, DivMod, IntVal
from hugr.std.logic import Not

from .conftest import validate


def test_stable_indices():
    h = Hugr(ops.DFG([]))

    nodes = [h.add_node(Not, num_outs=1) for _ in range(3)]
    assert len(h) == 4
    assert list(iter(h)) == [Node(i) for i in range(4)]
    assert all(data is not None for node, data in h.nodes())

    assert len(list(nodes[0].outputs())) == 1
    assert list(nodes[0]) == list(nodes[0].outputs())

    h.add_link(nodes[0].out(0), nodes[1].inp(0))
    assert h.children() == nodes

    assert h.num_outgoing(nodes[0]) == 1
    assert h.num_incoming(nodes[1]) == 1

    assert h.delete_node(nodes[1]) is not None
    assert h._nodes[nodes[1].idx] is None
    assert nodes[1] not in h.children(h.root)

    assert len(h) == 3
    assert len(h._nodes) == 4
    assert h._free_nodes == [nodes[1]]

    assert h.num_outgoing(nodes[0]) == 0
    assert h.num_incoming(nodes[1]) == 0

    with pytest.raises(KeyError):
        _ = h[nodes[1]]
    with pytest.raises(KeyError):
        _ = h[Node(46)]

    new_n = h.add_node(Not)
    assert new_n == nodes[1]

    assert len(h) == 4
    assert h._free_nodes == []
    assert list(iter(h)) == [Node(i) for i in range(4)]
    assert all(data is not None for node, data in h.nodes())


def simple_id() -> Dfg:
    h = Dfg(tys.Qubit, tys.Qubit)
    a, b = h.inputs()
    h.set_outputs(a, b)
    return h


def test_simple_id(snapshot):
    hugr = simple_id().hugr
    validate(hugr, snap=snapshot)


def test_metadata(snapshot):
    h = Dfg(tys.Bool)
    h.metadata["name"] = "simple_id"

    (b,) = h.inputs()
    b = h.add_op(Not, b, metadata={"name": "not"})

    h.set_outputs(b)
    validate(h.hugr, snap=snapshot)


def test_multiport(snapshot):
    h = Dfg(tys.Bool)
    (a,) = h.inputs()
    h.set_outputs(a, a)
    in_n, ou_n = h.input_node, h.output_node
    assert list(h.hugr.outgoing_links(in_n)) == [
        (in_n.out(0), [ou_n.inp(0), ou_n.inp(1)]),
    ]

    assert list(h.hugr.incoming_links(ou_n)) == [
        (ou_n.inp(0), [in_n.out(0)]),
        (ou_n.inp(1), [in_n.out(0)]),
    ]

    assert list(h.hugr.linked_ports(in_n.out(0))) == [
        ou_n.inp(0),
        ou_n.inp(1),
    ]

    assert list(h.hugr.linked_ports(ou_n.inp(0))) == [in_n.out(0)]
    validate(h.hugr, snap=snapshot)


def test_add_op(snapshot):
    h = Dfg(tys.Bool)
    (a,) = h.inputs()
    nt = h.add_op(Not, a)
    h.set_outputs(nt)

    validate(h.hugr, snap=snapshot)


def test_tuple(snapshot):
    row = [tys.Bool, tys.Qubit]
    h = Dfg(*row)
    a, b = h.inputs()
    t = h.add(ops.MakeTuple()(a, b))
    a, b = h.add(ops.UnpackTuple()(t))
    h.set_outputs(a, b)

    validate(h.hugr, snap=snapshot)

    h1 = Dfg(*row)
    a, b = h1.inputs()
    mt = h1.add_op(ops.MakeTuple(), a, b)
    a, b = h1.add_op(ops.UnpackTuple(), mt)[0, 1]
    h1.set_outputs(a, b)

    assert h.hugr._to_serial() == h1.hugr._to_serial()


def test_multi_out(snapshot):
    h = Dfg(INT_T, INT_T)
    a, b = h.inputs()
    a, b = h.add(DivMod(a, b))
    h.set_outputs(a, b)
    validate(h.hugr, snap=snapshot)


def test_insert():
    h1 = Dfg(tys.Bool)
    (a1,) = h1.inputs()
    nt = h1.add(Not(a1))
    h1.set_outputs(nt)

    assert len(h1.hugr) == 4

    new_h = Hugr(ops.DFG([]))
    mapping = h1.hugr.insert_hugr(new_h, h1.hugr.root)
    assert mapping == {new_h.root: Node(4)}


def test_insert_nested(snapshot):
    h1 = Dfg(tys.Bool)
    (a1,) = h1.inputs()
    nt = h1.add(Not(a1))
    h1.set_outputs(nt)

    h = Dfg(tys.Bool)
    (a,) = h.inputs()
    nested = h.insert_nested(h1, a)
    h.set_outputs(nested)
    assert len(h.hugr.children(nested)) == 3
    validate(h.hugr, snap=snapshot)


def test_build_nested(snapshot):
    h = Dfg(tys.Bool)
    (a,) = h.inputs()

    with h.add_nested(a) as nested:
        (a1,) = nested.inputs()
        nt = nested.add(Not(a1))
        nested.set_outputs(nt)

    assert len(h.hugr.children(nested)) == 3
    h.set_outputs(nested)

    validate(h.hugr, snap=snapshot)


def test_build_inter_graph(snapshot):
    h = Dfg(tys.Bool, tys.Bool)
    (a, b) = h.inputs()
    with h.add_nested() as nested:
        nt = nested.add(Not(a))
        nested.set_outputs(nt)

    h.set_outputs(nested, b)

    validate(h.hugr, snap=snapshot)

    assert _SubPort(h.input_node.out(-1)) in h.hugr._links
    assert h.hugr.num_outgoing(h.input_node) == 2  # doesn't count state order
    assert len(list(h.hugr.outgoing_order_links(h.input_node))) == 1
    assert len(list(h.hugr.incoming_order_links(nested))) == 1
    assert len(list(h.hugr.incoming_order_links(h.output_node))) == 0


def test_ancestral_sibling():
    h = Dfg(tys.Bool)
    (a,) = h.inputs()
    with h.add_nested() as nested:
        nt = nested.add(Not(a))

    assert _ancestral_sibling(h.hugr, h.input_node, nt) == nested.parent_node


@pytest.mark.parametrize(
    "val",
    [
        val.Function(simple_id().hugr),
        val.Sum(1, tys.Sum([[INT_T], [tys.Bool, INT_T]]), [val.TRUE, IntVal(34)]),
        val.Tuple(val.TRUE, IntVal(23)),
    ],
)
def test_vals(val: val.Value):
    d = Dfg()
    d.set_outputs(d.load(val))

    validate(d.hugr)


@pytest.mark.parametrize("direct_call", [True, False])
def test_poly_function(direct_call: bool) -> None:
    mod = Module()
    f_id = mod.declare_function(
        "id",
        tys.PolyFuncType(
            [tys.TypeTypeParam(tys.TypeBound.Any)],
            tys.FunctionType.endo([tys.Variable(0, tys.TypeBound.Any)]),
        ),
    )

    f_main = mod.define_main([tys.Qubit])
    q = f_main.input_node[0]
    # for now concrete instantiations have to be provided.
    instantiation = tys.FunctionType.endo([tys.Qubit])
    type_args = [tys.Qubit.type_arg()]
    if direct_call:
        with pytest.raises(NoConcreteFunc, match="Missing instantiation"):
            f_main.call(f_id, q)
        call = f_main.call(f_id, q, instantiation=instantiation, type_args=type_args)
    else:
        with pytest.raises(NoConcreteFunc, match="Missing instantiation"):
            f_main.load_function(f_id)
        load = f_main.load_function(
            f_id, instantiation=instantiation, type_args=type_args
        )
        call = f_main.add(ops.CallIndirect()(load, q))

    f_main.set_outputs(call)

    validate(mod.hugr)


@pytest.mark.parametrize("direct_call", [True, False])
def test_mono_function(direct_call: bool) -> None:
    mod = Module()
    f_id = mod.define_function("id", [tys.Qubit])
    f_id.set_outputs(f_id.input_node[0])

    f_main = mod.define_main([tys.Qubit])
    q = f_main.input_node[0]
    # monomorphic functions don't need instantiation specified
    if direct_call:
        call = f_main.call(f_id, q)
    else:
        load = f_main.load_function(f_id)
        call = f_main.add(ops.CallIndirect()(load, q))
    f_main.set_outputs(call)

    validate(mod.hugr)


def test_recursive_function(snapshot) -> None:
    mod = Module()

    f_recursive = mod.define_function("recurse", [tys.Qubit])
    f_recursive.declare_outputs([tys.Qubit])
    call = f_recursive.call(f_recursive, f_recursive.input_node[0])
    f_recursive.set_outputs(call)

    validate(mod.hugr, snap=snapshot)


def test_invalid_recursive_function() -> None:
    mod = Module()

    f_recursive = mod.define_function("recurse", [tys.Bool], [tys.Qubit])
    f_recursive.call(f_recursive, f_recursive.input_node[0])

    with pytest.raises(ValueError, match="The function has fixed output type"):
        f_recursive.set_outputs(f_recursive.input_node[0])


def test_higher_order() -> None:
    noop_fn = Dfg(tys.Qubit)
    noop_fn.set_outputs(noop_fn.add(ops.Noop()(noop_fn.input_node[0])))

    d = Dfg(tys.Qubit)
    (q,) = d.inputs()
    f_val = d.load(val.Function(noop_fn.hugr))
    call = d.add(ops.CallIndirect()(f_val, q))[0]
    d.set_outputs(call)

    validate(d.hugr)


def test_state_order() -> None:
    mod = Module()
    f_id = mod.define_function("id", [tys.Bool])
    f_id.set_outputs(f_id.input_node[0])

    f_main = mod.define_main([tys.Bool])
    b = f_main.input_node[0]
    call1 = f_main.call(f_id, b)
    f_main.add_state_order(call1, f_main.output_node)
    # implicit discard of bool to test state order port logic
    f_main.set_outputs()
    validate(mod.hugr)


def test_alias() -> None:
    mod = Module()
    _dfn = mod.add_alias_defn("my_int", INT_T)
    _dcl = mod.add_alias_decl("my_bool", tys.TypeBound.Copyable)

    validate(mod.hugr)


# https://github.com/CQCL/hugr/issues/1625
def test_dfg_unpack() -> None:
    dfg = Dfg(tys.Tuple(tys.Bool, tys.Bool))
    bool1, _unused_bool2 = dfg.add_op(ops.UnpackTuple(), *dfg.inputs())
    cond = dfg.add_conditional(bool1)
    with cond.add_case(0) as case:
        case.set_outputs(bool1)
    with cond.add_case(1) as case:
        case.set_outputs(bool1)
    dfg.set_outputs(*cond.outputs())

    validate(dfg.hugr)


def test_option() -> None:
    dfg = Dfg(tys.Bool)
    b = dfg.inputs()[0]

    dfg.add_op(ops.Some(tys.Bool), b)

    dfg.set_outputs(b)

    validate(dfg.hugr)



"""Generic serialization of HUGRs and Packages.

The format is designed to be extensible and backwards-compatible. It
consists of a header declaring the format used to encode the HUGR, followed
by the encoded HUGR itself.

## Payload formats

The envelope may encode the HUGR in different formats, listed in
:class:`hugr.envelope.EnvelopeFormat`. The payload may also be compressed with zstd.

Some formats can be represented as ASCII, as indicated by the
:meth:`hugr.envelope.EnvelopeFormat.ascii_printable` method. When this is the case, the
whole envelope can be stored in a string.

## Envelope header

The binary header format is 10 bytes, with the following fields:

| Field  | Size (bytes) | Description |
|--------|--------------|-------------|
| Magic  | 8            | :class:`hugr.envelope.MAGIC_NUMBERS` constant identifying the envelope format. |
| Format | 1            | :class:`hugr.envelope.EnvelopeFormat` describing the payload format. |
| Flags  | 1            | Additional configuration flags. |

Flags:

- Bit 0: Whether the payload is compressed with zstd.
- Bits 1-5: Reserved for future use.
- Bit 7,6: Constant "01" to make some headers ascii-printable.
"""  # noqa: E501

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, ClassVar

if TYPE_CHECKING:
    from hugr.package import Package

# This is a hard-coded magic number that identifies the start of a HUGR envelope.
MAGIC_NUMBERS = b"HUGRiHJv"


def make_envelope(package: Package, config: EnvelopeConfig) -> bytes:
    """Encode a HUGR package into an envelope, using the given configuration."""
    envelope = bytearray(config._make_header().to_bytes())

    if config.zstd is not None:
        msg = "Zstd compression is not supported yet."
        raise ValueError(msg)

    # Currently only uncompressed JSON is supported.
    match config.format:
        case EnvelopeFormat.JSON:
            json = package._to_serial().model_dump_json()
            # This introduces an extra encode/decode roundtrip when calling
            # `make_envelope_str`, but we prioritize speed for binary formats.
            envelope += json.encode("utf-8")
        case _:
            msg = f"Unsupported envelope format: {config.format}"
            raise ValueError(msg)

    return bytes(envelope)


def make_envelope_str(package: Package, config: EnvelopeConfig) -> str:
    """Encode a HUGR package into an envelope, using the given configuration."""
    if not config.format.ascii_printable():
        msg = "Only ascii-printable envelope formats can be encoded into a string."
        raise ValueError(msg)
    envelope = make_envelope(package, config)
    return envelope.decode("utf-8")


def read_envelope(envelope: bytes) -> Package:
    """Decode a HUGR package from an envelope."""
    import hugr._serialization.extension as ext_s

    header = EnvelopeHeader.from_bytes(envelope)
    payload = envelope[10:]

    if header.zstd:
        msg = "Zstd compression is not supported yet."
        raise ValueError(msg)

    match header.format:
        case EnvelopeFormat.JSON:
            return ext_s.Package.model_validate_json(payload).deserialize()
        case _:
            msg = f"Unsupported envelope format: {header.format}"
            raise ValueError(msg)


def read_envelope_str(envelope: str) -> Package:
    """Decode a HUGR package from an envelope."""
    return read_envelope(envelope.encode("utf-8"))


class EnvelopeFormat(Enum):
    """Format used to encode a HUGR envelope."""

    MODULE = 1
    """A capnp-encoded hugr-module."""
    MODULE_WITH_EXTS = 2
    """A capnp-encoded hugr-module, immediately followed by a json-encoded
    extension registry."""
    JSON = 63  # '?' in ASCII
    """A json-encoded hugr-package. This format is ASCII-printable."""

    def ascii_printable(self) -> bool:
        return self in {EnvelopeFormat.JSON}


@dataclass
class EnvelopeHeader:
    """Header of a HUGR envelope.

    See the module docstring for the binary format.

    Attributes:
        format: The format used to encode the HUGR.
        zstd: Whether the payload is compressed with zstd.
            The compression level is detected from the payload.
    """

    format: EnvelopeFormat
    zstd: bool = False

    def to_bytes(self) -> bytes:
        header_bytes = bytearray(MAGIC_NUMBERS)
        header_bytes.append(self.format.value)
        flags = 0b01000000
        if self.zstd:
            flags |= 0b00000001
        header_bytes.append(flags)
        return bytes(header_bytes)

    @staticmethod
    def from_bytes(data: bytes) -> EnvelopeHeader:
        if len(data) < 10:
            msg = (
                "Invalid HUGR envelope."
                + f"Expected at least 10 bytes, but got {len(data)} bytes."
            )
            raise ValueError(msg)
        if data[:8] != MAGIC_NUMBERS:
            msg = (
                "Invalid magic number for HUGR envelope."
                + f"Expected {MAGIC_NUMBERS!r} but got {data[:8]!r}"
            )
            raise ValueError(msg)

        format: EnvelopeFormat = EnvelopeFormat(data[8])

        flags = data[9]
        zstd = bool(flags & 0b00000001)

        return EnvelopeHeader(format=format, zstd=zstd)


@dataclass
class EnvelopeConfig:
    """Configuration for writing a HUGR envelope.

    Attributes:
        format: The format to use for encoding the HUGR.
        zstd: The compression level to use, or None for no compression.
            Use 0 for the default level.
    """

    format: EnvelopeFormat = EnvelopeFormat.JSON
    zstd: int | None = None

    TEXT: ClassVar[EnvelopeConfig]

    def _make_header(self) -> EnvelopeHeader:
        return EnvelopeHeader(format=self.format, zstd=self.zstd is not None)


# Set EnvelopeConfig's class variables.
# This can only be done _after_ the class is defined.
EnvelopeConfig.TEXT = EnvelopeConfig(format=EnvelopeFormat.JSON, zstd=None)



# Configuration file for the Sphinx documentation builder.  # noqa: INP001
# See https://www.sphinx-doc.org/en/master/usage/configuration.html


project = "HUGR Python"
copyright = "2024, Quantinuum"
author = "Quantinuum"

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
]

html_theme = "sphinx_book_theme"

html_title = "HUGR python package API documentation."

html_theme_options = {
    "repository_url": "https://github.com/CQCL/hugr",
    "use_repository_button": True,
    "navigation_with_keys": True,
    "logo": {
        "image_light": "_static/Quantinuum_logo_black.png",
        "image_dark": "_static/Quantinuum_logo_white.png",
    },
    "show_navbar_depth": 2,
}

html_static_path = ["../_static"]
html_css_files = ["custom.css"]

autosummary_generate = True

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "conftest.py"]


intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
}

html_show_sourcelink = False



"""Builder for HUGR datflow graphs."""

from __future__ import annotations

from contextlib import AbstractContextManager
from dataclasses import dataclass, field, replace
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
)

from typing_extensions import Self

from hugr import ops, tys, val
from hugr.build.base import ParentBuilder
from hugr.exceptions import NoSiblingAncestor
from hugr.hugr import Hugr

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from hugr.hugr.node_port import Node, OutPort, PortOffset, ToNode, Wire
    from hugr.tys import Type, TypeParam, TypeRow

    from .cfg import Cfg
    from .cond_loop import Conditional, If, TailLoop

OpVar = TypeVar("OpVar", bound=ops.Op)


@dataclass()
class DefinitionBuilder(Generic[OpVar]):
    """Base class for builders that can define functions, constants, and aliases.

    As this class may be a root node, it does not extend `ParentBuilder`.
    """

    hugr: Hugr[OpVar]

    def define_function(
        self,
        name: str,
        input_types: TypeRow,
        output_types: TypeRow | None = None,
        type_params: list[TypeParam] | None = None,
        parent: ToNode | None = None,
    ) -> Function:
        """Start building a function definition in the graph.

        Args:
            name: The name of the function.
            input_types: The input types for the function.
            output_types: The output types for the function.
                If not provided, it will be inferred after the function is built.
            type_params: The type parameters for the function, if polymorphic.
            parent: The parent node of the constant. Defaults to the root node.

        Returns:
            The new function builder.
        """
        parent_node = parent or self.hugr.root
        parent_op = ops.FuncDefn(name, input_types, type_params or [])
        func = Function.new_nested(parent_op, self.hugr, parent_node)
        if output_types is not None:
            func.declare_outputs(output_types)
        return func

    def add_const(self, value: val.Value, parent: ToNode | None = None) -> Node:
        """Add a static constant to the graph.

        Args:
            value: The constant value to add.
            parent: The parent node of the constant. Defaults to the root node.

        Returns:
            The node holding the :class:`Const <hugr.ops.Const>` operation.

        Example:
            >>> dfg = Dfg()
            >>> const_n = dfg.add_const(val.TRUE)
            >>> dfg.hugr[const_n].op
            Const(TRUE)
        """
        parent_node = parent or self.hugr.root
        return self.hugr.add_node(ops.Const(value), parent_node)

    def add_alias_defn(self, name: str, ty: Type, parent: ToNode | None = None) -> Node:
        """Add a type alias definition."""
        parent_node = parent or self.hugr.root
        return self.hugr.add_node(ops.AliasDefn(name, ty), parent_node)


DP = TypeVar("DP", bound=ops.DfParentOp)


@dataclass()
class DfBase(ParentBuilder[DP], DefinitionBuilder, AbstractContextManager):
    """Base class for dataflow graph builders.

    Args:
        parent_op: The parent operation of the dataflow graph.
    """

    #: The Hugr instance that the builder is using.
    hugr: Hugr = field(repr=False)
    #: The parent node of the dataflow graph.
    parent_node: Node
    #: The input node of the dataflow graph.
    input_node: Node = field(repr=False)
    #: The output node of the dataflow graph.
    output_node: Node = field(repr=False)

    def __init__(self, parent_op: DP) -> None:
        self.hugr = Hugr(parent_op)
        self.parent_node = self.hugr.root
        self._init_io_nodes(parent_op)

    def _init_io_nodes(self, parent_op: DP):
        inputs = parent_op._inputs()

        self.input_node = self.hugr.add_node(
            ops.Input(inputs), self.parent_node, len(inputs)
        )
        self.output_node = self.hugr.add_node(ops.Output(), self.parent_node)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args) -> None:
        return None

    @classmethod
    def new_nested(
        cls, parent_op: DP, hugr: Hugr, parent: ToNode | None = None
    ) -> Self:
        """Start building a dataflow graph nested inside a larger HUGR.

        Args:
            parent_op: The parent operation of the new dataflow graph.
            hugr: The host HUGR instance to build the dataflow graph in.
            parent: Parent of new dataflow graph's root node: defaults to the
            host HUGR root.

        Example:
            >>> hugr = Hugr()
            >>> dfg = Dfg.new_nested(ops.DFG([]), hugr)
            >>> dfg.parent_node
            Node(1)
        """
        new = cls.__new__(cls)

        new.hugr = hugr
        new.parent_node = hugr.add_node(parent_op, parent or hugr.root)
        new._init_io_nodes(parent_op)
        return new

    def _input_op(self) -> ops.Input:
        return self.hugr._get_typed_op(self.input_node, ops.Input)

    def _output_op(self) -> ops.Output:
        return self.hugr._get_typed_op(self.output_node, ops.Output)

    def inputs(self) -> list[OutPort]:
        """List all incoming wires (output ports of the input node).

        Example:
            >>> dfg = Dfg(tys.Bool)
            >>> dfg.inputs()
            [OutPort(Node(1), 0)]
        """
        return [self.input_node.out(i) for i in range(len(self._input_op().types))]

    def add_op(
        self, op: ops.DataflowOp, /, *args: Wire, metadata: dict[str, Any] | None = None
    ) -> Node:
        """Add a dataflow operation to the graph, wiring in input ports.

        Args:
            op: The operation to add.
            args: The input wires to the operation.
            metadata: Metadata to attach to the function definition. Defaults to None.

        Returns:
            The node holding the new operation.

        Example:
            >>> dfg = Dfg(tys.Bool)
            >>> dfg.add_op(ops.Noop(), dfg.inputs()[0])
            Node(3)
        """
        new_n = self.hugr.add_node(op, self.parent_node, metadata=metadata)
        self._wire_up(new_n, args)

        return replace(new_n, _num_out_ports=op.num_out)

    def add(self, com: ops.Command, *, metadata: dict[str, Any] | None = None) -> Node:
        """Add a command (holding a dataflow operation and the incoming wires)
        to the graph.

        Args:
            com: The command to add.
            metadata: Metadata to attach to the function definition. Defaults to None.

        Example:
            >>> dfg = Dfg(tys.Bool)
            >>> (i,) = dfg.inputs()
            >>> dfg.add(ops.Noop()(i))
            Node(3)

        """

        def raise_no_ints():
            error_message = "Command used with Dfg must hold Wire, not integer indices."
            raise ValueError(error_message)

        wires = (
            (w if not isinstance(w, int) else raise_no_ints()) for w in com.incoming
        )
        return self.add_op(com.op, *wires, metadata=metadata)

    def extend(self, *coms: ops.Command) -> list[Node]:
        """Add a series of commands to the DFG.

        Shorthand for calling :meth:`add` on each command in `coms`.

        Args:
            coms: Commands to add.

        Returns:
            List of the new nodes in the same order as the commands.

        Raises:
            IndexError: If any input index is not a tracked wire.

        Examples:
            >>> dfg = Dfg(tys.Bool, tys.Unit)
            >>> (b, u) = dfg.inputs()
            >>> dfg.extend(ops.Noop()(b), ops.Noop()(u))
            [Node(3), Node(4)]
        """
        return [self.add(com) for com in coms]

    def _insert_nested_impl(self, builder: ParentBuilder, *args: Wire) -> Node:
        mapping = self.hugr.insert_hugr(builder.hugr, self.parent_node)
        self._wire_up(mapping[builder.parent_node], args)
        return mapping[builder.parent_node]

    def insert_nested(self, dfg: Dfg, *args: Wire) -> Node:
        """Insert a nested dataflow graph into the current graph, wiring in the
        inputs.

        Args:
            dfg: The dataflow graph to insert.
            args: The input wires to the graph.

        Returns:
            The root node of the inserted graph.

        Example:
            >>> dfg = Dfg(tys.Bool)
            >>> dfg2 = Dfg(tys.Bool)
            >>> dfg.insert_nested(dfg2, dfg.inputs()[0])
            Node(3)
        """
        return self._insert_nested_impl(dfg, *args)

    def add_nested(
        self,
        *args: Wire,
    ) -> Dfg:
        """Start building a nested dataflow graph.

        Args:
            args: The input wires to the nested DFG.

        Returns:
            Builder for new nested dataflow graph.

        Example:
            >>> dfg = Dfg(tys.Bool)
            >>> with dfg.add_nested(dfg.inputs()[0]) as dfg2:\
                   dfg2.parent_node
            Node(3)
        """
        from .dfg import Dfg

        parent_op = ops.DFG(self._wire_types(args))
        dfg = Dfg.new_nested(parent_op, self.hugr, self.parent_node)
        self._wire_up(dfg.parent_node, args)
        return dfg

    def _wire_types(self, args: Iterable[Wire]) -> tys.TypeRow:
        return [self._get_dataflow_type(w) for w in args]

    def add_cfg(
        self,
        *args: Wire,
    ) -> Cfg:
        """Start building a new CFG nested inside the current dataflow graph.

        Args:
            args: The input wires to the new CFG.

        Returns:
            Builder for new nested CFG.

        Example:
            >>> dfg = Dfg(tys.Bool)
            >>> with dfg.add_cfg(dfg.inputs()[0]) as cfg:\
                    cfg.parent_op
            CFG(inputs=[Bool])
        """
        from .cfg import Cfg

        cfg = Cfg.new_nested(self._wire_types(args), self.hugr, self.parent_node)
        self._wire_up(cfg.parent_node, args)
        return cfg

    def insert_cfg(self, cfg: Cfg, *args: Wire) -> Node:
        """Insert a CFG into the current dataflow graph, wiring in the inputs.

        Args:
            cfg: The CFG to insert.
            args: The input wires to the CFG.

        Returns:
            The root node of the inserted CFG.

        Example:
            >>> from hugr.cfg import Cfg
            >>> dfg = Dfg(tys.Bool)
            >>> cfg = Cfg(tys.Bool)
            >>> dfg.insert_cfg(cfg, dfg.inputs()[0])
            Node(3)
        """
        return self._insert_nested_impl(cfg, *args)

    def add_conditional(self, cond_wire: Wire, *args: Wire) -> Conditional:
        """Start building a new conditional nested inside the current dataflow
        graph.

        Args:
            cond_wire: The wire holding the value (of Sum type) to branch the
            conditional on.
            args: Remaining input wires to the conditional.

        Returns:
            Builder for new nested conditional.

        Example:
            >>> dfg = Dfg(tys.Bool, tys.Unit)
            >>> (cond, unit) = dfg.inputs()
            >>> cond = dfg.add_conditional(cond, unit)
            >>> cond.parent_node
            Node(3)
        """
        from .cond_loop import Conditional

        args = (cond_wire, *args)
        (sum_, other_inputs) = tys.get_first_sum(self._wire_types(args))
        cond_wire = Conditional.new_nested(
            sum_, other_inputs, self.hugr, self.parent_node
        )
        self._wire_up(cond_wire.parent_node, args)
        return cond_wire

    def insert_conditional(
        self, cond: Conditional, cond_wire: Wire, *args: Wire
    ) -> Node:
        """Insert a conditional into the current dataflow graph, wiring in the
        inputs.

        Args:
            cond: The conditional to insert.
            cond_wire: The wire holding the value (of Sum type)
              to branch the Conditional on.
            args: Remaining input wires to the conditional.

        Returns:
            The root node of the inserted conditional.

        Example:
            >>> from hugr.cond_loop import Conditional
            >>> cond = Conditional(tys.Bool, [])
            >>> dfg = Dfg(tys.Bool)
            >>> cond_n = dfg.insert_conditional(cond, dfg.inputs()[0])
            >>> dfg.hugr[cond_n].op
            Conditional(sum_ty=Bool, other_inputs=[])
        """
        return self._insert_nested_impl(cond, *(cond_wire, *args))

    def add_if(self, cond_wire: Wire, *args: Wire) -> If:
        """Start building a new if block nested inside the current dataflow
        graph.

        Args:
            cond_wire: The wire holding the Bool value to branch the If on.
            args: Remaining input wires to the If (and subsequent Else).

        Returns:
            Builder for new nested If.

        Example:
            >>> dfg = Dfg(tys.Bool)
            >>> (cond,) = dfg.inputs()
            >>> if_ = dfg.add_if(cond, cond)
            >>> if_.parent_op
            Case(inputs=[Bool])
        """
        from .cond_loop import If

        conditional = self.add_conditional(cond_wire, *args)
        return If(conditional.add_case(1))

    def add_tail_loop(
        self, just_inputs: Sequence[Wire], rest: Sequence[Wire]
    ) -> TailLoop:
        """Start building a new tail loop nested inside the current dataflow
        graph.

        Args:
            just_inputs: input wires for types that are only inputs to the loop body.
            rest: input wires for types that are inputs and outputs of the loop
            body.

        Returns:
            Builder for new nested TailLoop.

        Example:
            >>> dfg = Dfg(tys.Bool)
            >>> (cond,) = dfg.inputs()
            >>> tl = dfg.add_tail_loop([cond], [cond])
            >>> tl.parent_op
            TailLoop(just_inputs=[Bool], rest=[Bool])
        """
        from .cond_loop import TailLoop

        just_input_types = self._wire_types(just_inputs)
        rest_types = self._wire_types(rest)
        parent_op = ops.TailLoop(just_input_types, rest_types)
        tl = TailLoop.new_nested(parent_op, self.hugr, self.parent_node)
        self._wire_up(tl.parent_node, (*just_inputs, *rest))
        return tl

    def insert_tail_loop(
        self, tl: TailLoop, just_inputs: Sequence[Wire], rest: Sequence[Wire]
    ) -> Node:
        """Insert a tail loop into the current dataflow graph, wiring in the
        inputs.

        Args:
            tl: The tail loop to insert.
            just_inputs: input wires for types that are only inputs to the loop body.
            rest: input wires for types that are inputs and outputs of the loop
            body.

        Returns:
            The root node of the inserted tail loop.

        Example:
            >>> from hugr.cond_loop import TailLoop
            >>> tl = TailLoop([tys.Bool], [tys.Bool])
            >>> dfg = Dfg(tys.Bool)
            >>> (b,) = dfg.inputs()
            >>> tl_n = dfg.insert_tail_loop(tl, [b], [b])
            >>> dfg.hugr[tl_n].op
            TailLoop(just_inputs=[Bool], rest=[Bool])
        """
        return self._insert_nested_impl(tl, *(*just_inputs, *rest))

    def set_outputs(self, *args: Wire) -> None:
        """Set the outputs of the dataflow graph.
        Connects wires to the output node.

        Args:
            args: Wires to connect to the output node.

        Example:
            >>> dfg = Dfg(tys.Bool)
            >>> dfg.set_outputs(dfg.inputs()[0]) # connect input to output
        """
        self._wire_up(self.output_node, args)
        self.parent_op._set_out_types(self._output_op().types)

    def _set_parent_output_count(self, count: int) -> None:
        """Set the final number of output ports on the parent operation.

        Args:
            count: The number of output ports.

        Example:
            >>> dfg = Dfg(tys.Bool)
            >>> dfg._set_parent_output_count(2)
        """
        self.parent_node = self.hugr._update_node_outs(self.parent_node, count)

    def add_state_order(self, src: Node, dst: Node) -> None:
        """Add a state order link between two nodes.

        Args:
            src: The source node.
            dst: The destination node.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.add_state_order(df.input_node, df.output_node)
            >>> list(df.hugr.outgoing_order_links(df.input_node))
            [Node(2)]
        """
        # adds edge to the right of all existing edges
        self.hugr.add_order_link(src, dst)

    def load(
        self, const: ToNode | val.Value, const_parent: ToNode | None = None
    ) -> Node:
        """Load a constant into the graph as a dataflow value.

        Args:
            const: The constant to load, either a Value that will be added as a
                child Const node then loaded, or a node corresponding to an
                existing Const.
            const_parent: If `const` is a Value, the parent node for the new
                constant definition. Defaults to the current dataflow container.

        Returns:
            The node holding the :class:`LoadConst <hugr.ops.LoadConst>`
            operation.

        Example:
            >>> dfg = Dfg()
            >>> const_n = dfg.load(val.TRUE)
            >>> len(dfg.hugr) # parent, input, output, const, load
            5
            >>> dfg.hugr[const_n].op
            LoadConst(Bool)
        """
        if isinstance(const, val.Value):
            const_parent = const_parent or self.parent_node
            const = self.add_const(const, parent=const_parent)
        const_op = self.hugr._get_typed_op(const, ops.Const)
        load_op = ops.LoadConst(const_op.val.type_())

        load = self.add(load_op())
        self.hugr.add_link(const.out_port(), load.inp(0))

        return load

    def call(
        self,
        func: ToNode,
        *args: Wire,
        instantiation: tys.FunctionType | None = None,
        type_args: Sequence[tys.TypeArg] | None = None,
    ) -> Node:
        """Call a static function in the graph.
        See :class:`Call <hugr.ops.Call>` for more on how polymorphic functions
        are handled.

        Args:
            func: The node corresponding to the function definition/declaration to call.
            args: The input wires to the function call.
            instantiation: The concrete function type to call (needed if polymorphic).
            type_args: The type arguments for the function (needed if
            polymorphic).

        Returns:
            The node holding the :class:`Call <hugr.ops.Call>` operation.
        """
        signature = self._fn_sig(func)
        call_op = ops.Call(signature, instantiation, type_args)
        call_n = self.hugr.add_node(call_op, self.parent_node, call_op.num_out)
        self.hugr.add_link(func.out(0), call_n.inp(call_op._function_port_offset()))

        self._wire_up(call_n, args)

        return call_n

    def load_function(
        self,
        func: ToNode,
        instantiation: tys.FunctionType | None = None,
        type_args: Sequence[tys.TypeArg] | None = None,
    ) -> Node:
        """Load a static function into the graph as a higher-order value.

        Args:
            func: The node corresponding to the function definition/declaration to load.
            instantiation: The concrete function type to load (needed if polymorphic).
            type_args: The type arguments for the function (needed if
            polymorphic).

        Returns:
            The node holding the :class:`LoadFunc <hugr.ops.LoadFunc>` operation.
        """
        signature = self._fn_sig(func)
        load_op = ops.LoadFunc(signature, instantiation, type_args)
        load_n = self.hugr.add_node(load_op, self.parent_node)
        self.hugr.add_link(func.out(0), load_n.inp(0))

        return load_n

    def _fn_sig(self, func: ToNode) -> tys.PolyFuncType:
        f_op = self.hugr[func]
        f_kind = f_op.op.port_kind(func.out(0))
        match f_kind:
            case tys.FunctionKind(sig):
                signature = sig
            case _:
                msg = "Expected 'func' to be a function"
                raise ValueError(msg)
        return signature

    def _wire_up(self, node: Node, ports: Iterable[Wire]) -> tys.TypeRow:
        tys = [self._wire_up_port(node, i, p) for i, p in enumerate(ports)]
        if isinstance(op := self.hugr[node].op, ops._PartialOp):
            op._set_in_types(tys)
            if isinstance(op, ops.DataflowOp):
                # Update the node's input and output port count
                sig = op.outer_signature()
                self.hugr._update_port_count(
                    node, num_inps=len(sig.input), num_outs=len(sig.output)
                )
        return tys

    def _get_dataflow_type(self, wire: Wire) -> tys.Type:
        port = wire.out_port()
        ty = self.hugr.port_type(port)
        if ty is None:
            msg = f"Port {port} is not a dataflow port."
            raise ValueError(msg)
        return ty

    def _wire_up_port(self, node: Node, offset: PortOffset, p: Wire) -> tys.Type:
        src = p.out_port()
        node_ancestor = _ancestral_sibling(self.hugr, src.node, node)
        if node_ancestor is None:
            raise NoSiblingAncestor(src.node.idx, node.idx)
        if node_ancestor != node:
            self.add_state_order(src.node, node_ancestor)
        self.hugr.add_link(src, node.inp(offset))
        return self._get_dataflow_type(src)


class Dfg(DfBase[ops.DFG]):
    """Builder for a simple nested Dataflow graph, with root node of type
    :class:`DFG <hugr.ops.DFG>`.

    Args:
        input_types: The input types of the the dataflow graph. Output types are
        calculated by propagating types through the graph.

    Example:
        >>> dfg = Dfg(tys.Bool)
        >>> dfg.parent_op
        DFG(inputs=[Bool])
    """

    def __init__(self, *input_types: tys.Type) -> None:
        parent_op = ops.DFG(list(input_types), None)
        super().__init__(parent_op)

    def set_outputs(self, *outputs: Wire) -> None:
        super().set_outputs(*outputs)
        self._set_parent_output_count(len(outputs))


def _ancestral_sibling(h: Hugr, src: Node, tgt: Node) -> Node | None:
    """Find the ancestor of `tgt` that is a sibling of `src`, if one exists."""
    src_parent = h[src].parent

    while (tgt_parent := h[tgt].parent) is not None:
        if tgt_parent == src_parent:
            return tgt
        tgt = tgt_parent

    return None


@dataclass
class Function(DfBase[ops.FuncDefn]):
    """Build a function definition as a HUGR dataflow graph.

    Args:
        name: The name of the function.
        input_types: The input types for the function (output types are
        computed by propagating types from input node through the graph).
        type_params: The type parameters for the function, if polymorphic.

    Examples:
        >>> f = Function("f", [tys.Bool])
        >>> f.parent_op
        FuncDefn(f_name='f', inputs=[Bool], params=[])
    """

    def __init__(
        self,
        name: str,
        input_types: TypeRow,
        type_params: list[TypeParam] | None = None,
    ) -> None:
        root_op = ops.FuncDefn(name, input_types, type_params or [])
        super().__init__(root_op)

    def declare_outputs(self, output_types: TypeRow) -> None:
        """Declare the output types of the function.

        This is required when calling a function which hasn't been completely
        defined yet. The wires passed to :meth:`set_outputs` must match the
        declared output types.
        """
        self._set_parent_output_count(len(output_types))
        self.parent_op._set_out_types(output_types)

    def set_outputs(self, *args: Wire) -> None:
        """Set the outputs of the dataflow graph.
        Connects wires to the output node.

        If :meth:`declare_outputs` has been called, the wire types must match
        the declared output types.

        Args:
            args: Wires to connect to the output node.

        Example:
            >>> dfg = Dfg(tys.Bool)
            >>> dfg.set_outputs(dfg.inputs()[0]) # connect input to output
        """
        if self.parent_op._outputs is not None:
            arg_types = [self._get_dataflow_type(w) for w in args]
            if arg_types != self.parent_op._outputs:
                error_message = (
                    f"The function has fixed output type {self.parent_op._outputs}, "
                    f"but was given output wires with types {arg_types}."
                )
                raise ValueError(error_message)

        super().set_outputs(*args)



"""Core data structures for HUGR."""

from __future__ import annotations

import json
from collections.abc import Iterable, Iterator, Mapping
from dataclasses import dataclass, field, replace
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
    cast,
    overload,
)

from hugr._serialization.ops import OpType as SerialOp
from hugr._serialization.serial_hugr import SerialHugr
from hugr.exceptions import ParentBeforeChild
from hugr.ops import Call, Const, Custom, DataflowOp, Module, Op
from hugr.tys import Kind, Type, ValueKind
from hugr.utils import BiMap
from hugr.val import Value

from .node_port import (
    Direction,
    InPort,
    Node,
    NodeIdx,
    OutPort,
    PortOffset,
    ToNode,
    _SubPort,
)

if TYPE_CHECKING:
    import graphviz as gv  # type: ignore[import-untyped]

    from hugr import ext
    from hugr.val import Value

    from .render import RenderConfig


@dataclass()
class NodeData:
    """Node weights in HUGR graph. Defined by an operation and parent node."""

    #: The operation of the node.
    op: Op
    #: The parent node, or None for the root node.
    parent: Node | None
    _num_inps: int = field(default=0, repr=False)
    _num_outs: int = field(default=0, repr=False)
    children: list[Node] = field(default_factory=list, repr=False)
    metadata: dict[str, Any] = field(default_factory=dict)

    def _to_serial(self, node: Node) -> SerialOp:
        o = self.op._to_serial(self.parent if self.parent else node)

        return SerialOp(root=o)  # type: ignore[arg-type]


_SO = _SubPort[OutPort]
_SI = _SubPort[InPort]

P = TypeVar("P", InPort, OutPort)
K = TypeVar("K", InPort, OutPort)
OpVar = TypeVar("OpVar", bound=Op)
OpVarCov = TypeVar("OpVarCov", bound=Op, covariant=True)


@dataclass()
class Hugr(Mapping[Node, NodeData], Generic[OpVarCov]):
    """The core HUGR datastructure.

    Args:
        root_op: The operation for the root node. Defaults to a Module.

    Examples:
        >>> h = Hugr()
        >>> h.root_op()
        Module()
        >>> h[h.root].op
        Module()
    """

    #: Root node of the HUGR.
    root: Node
    # List of nodes, with None for deleted nodes.
    _nodes: list[NodeData | None]
    # Bidirectional map of links between ports.
    _links: BiMap[_SO, _SI]
    # List of free node indices, populated when nodes are deleted.
    _free_nodes: list[Node]

    def __init__(self, root_op: OpVarCov | None = None) -> None:
        self._free_nodes = []
        self._links = BiMap()
        self._nodes = []
        self.root = self._add_node(root_op or Module(), None, 0)

    def __getitem__(self, key: ToNode) -> NodeData:
        key = key.to_node()
        try:
            n = self._nodes[key.idx]
        except IndexError:
            n = None
        if n is None:
            raise KeyError(key)
        return n

    def __iter__(self) -> Iterator[Node]:
        return (
            Node(idx, data.metadata)
            for idx, data in enumerate(self._nodes)
            if data is not None
        )

    def __len__(self) -> int:
        return self.num_nodes()

    def _get_typed_op(self, node: ToNode, cl: type[OpVar]) -> OpVar:
        op = self[node].op
        assert isinstance(op, cl)
        return op

    def nodes(self) -> Iterable[tuple[Node, NodeData]]:
        """Iterator over nodes of the hugr and their data."""
        return self.items()

    def links(self) -> Iterator[tuple[OutPort, InPort]]:
        """Iterator over all the links in the HUGR.

        Returns:
            Iterator of pairs of outgoing port and the incoming ports.
        """
        return ((src.port, tgt.port) for src, tgt in self._links.items())

    def children(self, node: ToNode | None = None) -> list[Node]:
        """The child nodes of a given `node`.

        Args:
            node: Parent node. Defaults to the HUGR root.

        Returns:
            List of child nodes.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_node(ops.Const(val.TRUE))
            >>> h.children(h.root)
            [Node(1)]
        """
        node = node or self.root
        return self[node].children

    def _add_node(
        self,
        op: Op,
        parent: ToNode | None = None,
        num_outs: int | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> Node:
        parent = parent.to_node() if parent else None
        node_data = NodeData(op, parent, metadata=metadata or {})

        if self._free_nodes:
            node = self._free_nodes.pop()
            self._nodes[node.idx] = node_data
        else:
            node = Node(len(self._nodes), {})
            self._nodes.append(node_data)
        node = replace(node, _num_out_ports=num_outs, _metadata=node_data.metadata)
        if parent:
            self[parent].children.append(node)

        self._update_node_outs(node, num_outs)
        return node

    def _update_node_outs(self, node: Node, num_outs: int | None) -> Node:
        """Update the number of outgoing ports for a node.

        Returns:
            The updated node.
        """
        return self._update_port_count(node, num_outs=num_outs)

    def _update_port_count(
        self, node: Node, *, num_inps: int | None = None, num_outs: int | None
    ) -> Node:
        """Update the number of incoming and outgoing ports for a node.

        If `num_inps` or `num_outs` is None, the corresponding count is not updated.

        Returns:
            The updated node.
        """
        if num_inps is None and num_outs is None:
            return node

        if num_inps is not None:
            self[node]._num_inps = num_inps
        if num_outs is not None:
            self[node]._num_outs = num_outs
            node = replace(node, _num_out_ports=num_outs)
            parent = self[node].parent
            if parent is not None:
                pos = self[parent].children.index(node)
                self[parent].children[pos] = node

        return node

    def add_node(
        self,
        op: Op,
        parent: ToNode | None = None,
        num_outs: int | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> Node:
        """Add a node to the HUGR.

        Args:
            op: Operation of the node.
            parent: Parent node of added node. Defaults to HUGR root if None.
            num_outs: Number of output ports expected for this node. Defaults to None.
            metadata: A dictionary of metadata to associate with the node.
                Defaults to None.

        Returns:
            Handle to the added node.
        """
        parent = parent or self.root
        return self._add_node(op, parent, num_outs, metadata)

    def add_const(
        self,
        value: Value,
        parent: ToNode | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> Node:
        """Add a constant node to the HUGR.

        Args:
            value: Value of the constant.
            parent: Parent node of added node. Defaults to HUGR root if None.
            metadata: A dictionary of metadata to associate with the node.
                Defaults to None.

        Returns:
            Handle to the added node.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> h[n].op
            Const(TRUE)
        """
        return self.add_node(Const(value), parent, metadata=metadata)

    def delete_node(self, node: ToNode) -> NodeData | None:
        """Delete a node from the HUGR.

        Args:
            node: Node to delete.

        Returns:
            The deleted node data, or None if the node was not found.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> deleted = h.delete_node(n)
            >>> deleted.op
            Const(TRUE)
            >>> len(h)
            1
        """
        node = node.to_node()
        parent = self[node].parent
        if parent:
            self[parent].children.remove(node)
        for inp, _ in self.incoming_links(node):
            self._links.delete_right(_SubPort(inp))
        for out, _ in self.outgoing_links(node):
            self._links.delete_left(_SubPort(out))

        weight, self._nodes[node.idx] = self._nodes[node.idx], None

        # Free up the metadata dictionary
        node = replace(node, _metadata={})

        self._free_nodes.append(node)
        return weight

    def _unused_sub_offset(self, port: P) -> _SubPort[P]:
        d: dict[_SO, _SI] | dict[_SI, _SO]
        match port:
            case OutPort(_):
                d = self._links.fwd
            case InPort(_):
                d = self._links.bck
        sub_port = _SubPort(port)
        while sub_port in d:
            sub_port = sub_port.next_sub_offset()
        return sub_port

    def has_link(self, src: OutPort, dst: InPort) -> bool:
        """Check if there is a link between two ports.

        Args:
            src: Source port.
            dst: Destination port.

        Returns:
            True if there is a link, False otherwise.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.has_link(df.input_node.out(0), df.output_node.inp(0))
            True
        """
        return dst in self.linked_ports(src)

    def add_link(self, src: OutPort, dst: InPort) -> None:
        """Add a link (edge) between two nodes to the HUGR,
          from an outgoing port to an incoming port.

        Args:
            src: Source port.
            dst: Destination port.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> list(df.hugr.linked_ports(df.input_node[0]))
            [InPort(Node(2), 0)]
        """
        src_sub = self._unused_sub_offset(src)
        dst_sub = self._unused_sub_offset(dst)
        # if self._links.get_left(dst_sub) is not None:
        #     dst = replace(dst, _sub_offset=dst._sub_offset + 1)
        self._links.insert_left(src_sub, dst_sub)

        self[src.node]._num_outs = max(self[src.node]._num_outs, src.offset + 1)
        self[dst.node]._num_inps = max(self[dst.node]._num_inps, dst.offset + 1)

    def add_order_link(self, src: ToNode, dst: ToNode) -> None:
        """Add a state order link between two nodes.

        Args:
            src: Source node.
            dst: Destination node.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_order_link(df.input_node, df.output_node)
            >>> list(df.hugr.outgoing_order_links(df.input_node))
            [Node(2)]
        """
        source = src.out(-1)
        target = dst.inp(-1)
        if not self.has_link(source, target):
            self.add_link(source, target)

    def delete_link(self, src: OutPort, dst: InPort) -> None:
        """Delete a link (edge) between two nodes from the HUGR.

        Args:
            src: Source port.
            dst: Destination port.
        """
        try:
            sub_offset = next(
                i for i, inp in enumerate(self.linked_ports(src)) if inp == dst
            )
            self._links.delete_left(_SubPort(src, sub_offset))
        except StopIteration:
            return
        # TODO make sure sub-offset is handled correctly

    def root_op(self) -> OpVarCov:
        """The operation of the root node.

        Examples:
            >>> h = Hugr()
            >>> h.root_op()
            Module()
        """
        return cast(OpVarCov, self[self.root].op)

    def num_nodes(self) -> int:
        """The number of nodes in the HUGR.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> h.num_nodes()
            2
        """
        return len(self._nodes) - len(self._free_nodes)

    def num_ports(self, node: ToNode, direction: Direction) -> int:
        """The number of ports of a node in a given direction.
        Not necessarily the number of connected ports - if port `i` is
        connected, then all ports `0..i` are assumed to exist.

        Args:
            node: Node to query.
            direction: Direction of ports to count.

        Examples:
            >>> h = Hugr()
            >>> n1 = h.add_const(val.TRUE)
            >>> n2 = h.add_const(val.FALSE)
            >>> h.add_link(n1.out(0), n2.inp(2)) # not a valid link!
            >>> h.num_ports(n1, Direction.OUTGOING)
            1
            >>> h.num_ports(n2, Direction.INCOMING)
            3
        """
        return (
            self.num_in_ports(node)
            if direction == Direction.INCOMING
            else self.num_out_ports(node)
        )

    def num_in_ports(self, node: ToNode) -> int:
        """The number of incoming ports of a node. See :meth:`num_ports`."""
        return self[node]._num_inps

    def num_out_ports(self, node: ToNode) -> int:
        """The number of outgoing ports of a node. See :meth:`num_ports`."""
        return self[node]._num_outs

    def _linked_ports(
        self, port: P, links: dict[_SubPort[P], _SubPort[K]]
    ) -> Iterable[K]:
        sub_port = _SubPort(port)
        while sub_port in links:
            # sub offset not used in API
            yield links[sub_port].port
            sub_port = sub_port.next_sub_offset()

    @overload
    def linked_ports(self, port: OutPort) -> Iterable[InPort]: ...
    @overload
    def linked_ports(self, port: InPort) -> Iterable[OutPort]: ...
    def linked_ports(self, port: OutPort | InPort):
        """Return an iterable of In(Out)Ports linked to given Out(In)Port.

        Args:
            port: Given port.

        Returns:
            Iterator over linked ports.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.set_outputs(df.input_node[0])
            >>> list(df.hugr.linked_ports(df.input_node[0]))
            [InPort(Node(2), 0)]

        """
        match port:
            case OutPort(_):
                return self._linked_ports(port, self._links.fwd)
            case InPort(_):
                return self._linked_ports(port, self._links.bck)

    # TODO: single linked port

    def outgoing_order_links(self, node: ToNode) -> Iterable[Node]:
        """Iterator over nodes connected by an outgoing state order link from a
        given node.

        Args:
            node: Source node of state order link.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.add_state_order(df.input_node, df.output_node)
            >>> list(df.hugr.outgoing_order_links(df.input_node))
            [Node(2)]
        """
        return (p.node for p in self.linked_ports(node.out(-1)))

    def incoming_order_links(self, node: ToNode) -> Iterable[Node]:
        """Iterator over nodes connected by an incoming state order link to a
        given node.

        Args:
            node: Destination node of state order link.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.add_state_order(df.input_node, df.output_node)
            >>> list(df.hugr.incoming_order_links(df.output_node))
            [Node(1)]
        """
        return (p.node for p in self.linked_ports(node.inp(-1)))

    def _node_links(
        self, node: ToNode, links: dict[_SubPort[P], _SubPort[K]]
    ) -> Iterable[tuple[P, list[K]]]:
        try:
            direction = next(iter(links.keys())).port.direction
        except StopIteration:
            return
        # iterate over known offsets
        for offset in range(self.num_ports(node, direction)):
            port = cast(P, node.port(offset, direction))
            yield port, list(self._linked_ports(port, links))

    def outgoing_links(self, node: ToNode) -> Iterable[tuple[OutPort, list[InPort]]]:
        """Iterator over outgoing links from a given node.

        Args:
            node: Node to query.

        Returns:
            Iterator of pairs of outgoing port and the incoming ports connected
            to that port.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.outgoing_links(df.input_node))
            [(OutPort(Node(1), 0), [InPort(Node(2), 0), InPort(Node(2), 1)])]
        """
        return self._node_links(node, self._links.fwd)

    def incoming_links(self, node: ToNode) -> Iterable[tuple[InPort, list[OutPort]]]:
        """Iterator over incoming links to a given node.

        Args:
            node: Node to query.

        Returns:
            Iterator of pairs of incoming port and the outgoing ports connected
            to that port.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.incoming_links(df.output_node))
            [(InPort(Node(2), 0), [OutPort(Node(1), 0)]), (InPort(Node(2), 1), [OutPort(Node(1), 0)])]
        """  # noqa: E501
        return self._node_links(node, self._links.bck)

    def num_incoming(self, node: Node) -> int:
        """The number of incoming links to a `node`.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.num_incoming(df.output_node)
            1
        """
        return sum(1 for _ in self.incoming_links(node))

    def num_outgoing(self, node: ToNode) -> int:
        """The number of outgoing links from a `node`.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.num_outgoing(df.input_node)
            1
        """
        return sum(1 for _ in self.outgoing_links(node))

    # TODO: num_links and _linked_ports

    def port_kind(self, port: InPort | OutPort) -> Kind:
        """The kind of a `port`.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.port_kind(df.input_node.out(0))
            ValueKind(Bool)
        """
        return self[port.node].op.port_kind(port)

    def port_type(self, port: InPort | OutPort) -> Type | None:
        """The type of a `port`, if the kind is
        :class:`ValueKind <hugr.tys.ValueKind>`, else None.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.port_type(df.input_node.out(0))
            Bool
        """
        op = self[port.node].op
        if isinstance(op, DataflowOp):
            return op.port_type(port)
        if isinstance(op, Call) and isinstance(port, OutPort):
            kind = self.port_kind(port)
            if isinstance(kind, ValueKind):
                return kind.ty
        return None

    def insert_hugr(self, hugr: Hugr, parent: ToNode | None = None) -> dict[Node, Node]:
        """Insert a HUGR into this HUGR.

        Args:
            hugr: HUGR to insert.
            parent: Parent for root of inserted HUGR. Defaults to None.

        Returns:
            Mapping from node indices in inserted HUGR to their new indices
            in this HUGR.

        Examples:
            >>> d = dfg.Dfg()
            >>> h = Hugr()
            >>> h.insert_hugr(d.hugr)
            {Node(0): Node(1), Node(1): Node(2), Node(2): Node(3)}
        """
        mapping: dict[Node, Node] = {}

        for node, node_data in hugr.nodes():
            # relies on parents being inserted before any children
            try:
                node_parent = mapping[node_data.parent] if node_data.parent else parent
            except KeyError as e:
                raise ParentBeforeChild from e
            mapping[node] = self.add_node(
                node_data.op,
                node_parent,
                num_outs=node_data._num_outs,
                metadata=node_data.metadata,
            )

        for src, dst in hugr._links.items():
            self.add_link(
                mapping[src.port.node].out(src.port.offset),
                mapping[dst.port.node].inp(dst.port.offset),
            )
        return mapping

    def _to_serial(self) -> SerialHugr:
        """Serialize the HUGR."""
        node_it = (node for node in self._nodes if node is not None)

        def _serialize_link(
            link: tuple[_SO, _SI],
        ) -> tuple[tuple[NodeIdx, PortOffset], tuple[NodeIdx, PortOffset]]:
            src, dst = link
            s, d = self._constrain_offset(src.port), self._constrain_offset(dst.port)
            return (src.port.node.idx, s), (dst.port.node.idx, d)

        return SerialHugr(
            # non contiguous indices will be erased
            nodes=[node._to_serial(Node(idx, {})) for idx, node in enumerate(node_it)],
            edges=[_serialize_link(link) for link in self._links.items()],
            metadata=[node.metadata if node.metadata else None for node in node_it],
        )

    def _constrain_offset(self, p: P) -> PortOffset:
        # An offset of -1 is a special case, indicating an order edge,
        # not counted in the number of ports.
        if p.offset < 0:
            assert p.offset == -1, "Only order edges are allowed with offset < 0"
            offset = self.num_ports(p.node, p.direction)
        else:
            offset = p.offset

        return offset

    def resolve_extensions(self, registry: ext.ExtensionRegistry) -> Hugr:
        """Resolve extension types and operations in the HUGR by matching them to
        extensions in the registry.
        """
        for node in self:
            op = self[node].op
            if isinstance(op, Custom):
                self[node].op = op.resolve(registry)
        return self

    @classmethod
    def _from_serial(cls, serial: SerialHugr) -> Hugr:
        """Load a HUGR from a serialized form."""
        assert serial.nodes, "Empty Hugr is invalid"

        hugr = Hugr.__new__(Hugr)
        hugr._nodes = []
        hugr._links = BiMap()
        hugr._free_nodes = []
        hugr.root = Node(0)

        def get_meta(idx: int) -> dict[str, Any]:
            if not serial.metadata:
                return {}
            if idx < len(serial.metadata):
                return serial.metadata[idx] or {}
            return {}

        for idx, serial_node in enumerate(serial.nodes):
            node_meta = get_meta(idx)
            parent: Node | None = Node(serial_node.root.parent)
            if serial_node.root.parent == idx:
                hugr.root = Node(idx, _metadata=node_meta)
                parent = None

            serial_node.root.parent = -1
            n = hugr._add_node(
                serial_node.root.deserialize(), parent, metadata=node_meta
            )
            assert n.idx == idx, "Nodes should be added contiguously"

        for (src_node, src_offset), (dst_node, dst_offset) in serial.edges:
            if src_offset is None or dst_offset is None:
                continue
            hugr.add_link(
                Node(src_node, _metadata=get_meta(src_node)).out(src_offset),
                Node(dst_node, _metadata=get_meta(dst_node)).inp(dst_offset),
            )

        return hugr

    def to_json(self) -> str:
        """Serialize the HUGR to a JSON string.

        For most use cases, it is recommended to store a HUGR package instead.
        See :meth:`hugr.package.Package.to_bytes`.
        """
        return self._to_serial().to_json()

    @classmethod
    def load_json(cls, json_str: str) -> Hugr:
        """Deserialize a JSON string into a HUGR.

        For most use cases, it is recommended to use package serialization instead.
        See :meth:`hugr.package.Package.from_bytes`.
        """
        json_dict = json.loads(json_str)
        serial = SerialHugr.load_json(json_dict)
        return cls._from_serial(serial)

    def render_dot(self, config: RenderConfig | None = None) -> gv.Digraph:
        """Render the HUGR to a graphviz Digraph.

        Args:
            config: Render configuration.

        Returns:
            The graphviz Digraph.
        """
        from .render import DotRenderer

        return DotRenderer(config).render(self)

    def store_dot(
        self, filename: str, format: str = "svg", config: RenderConfig | None = None
    ) -> None:
        """Render the HUGR to a graphviz dot file.

        Args:
            filename: The file to render to.
            format: The format used for rendering ('pdf', 'png', etc.).
                Defaults to SVG.
            config: Render configuration.
        """
        from .render import DotRenderer

        DotRenderer(config).store(self, filename=filename, format=format)



"""Builder classes for structured control flow
in HUGR graphs (Conditional, TailLoop).
"""

from __future__ import annotations

from contextlib import AbstractContextManager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from typing_extensions import Self

from hugr import ops
from hugr.build.base import ParentBuilder
from hugr.build.dfg import DfBase
from hugr.hugr.base import Hugr
from hugr.tys import Sum

if TYPE_CHECKING:
    from hugr.hugr.node_port import Node, ToNode, Wire
    from hugr.tys import TypeRow
import warnings


class Case(DfBase[ops.Case]):
    """Dataflow graph builder for a case in a conditional."""

    _parent_cond: Conditional | None = None

    def set_outputs(self, *outputs: Wire) -> None:
        super().set_outputs(*outputs)
        if self._parent_cond is not None:
            self._parent_cond._update_outputs(self._wire_types(outputs))


class ConditionalError(Exception):
    """Error building a :class:`Conditional`."""


@dataclass
class _IfElse(Case):
    def __init__(self, case: Case) -> None:
        self.hugr = case.hugr
        self.parent_node = case.parent_node
        self.input_node = case.input_node
        self.output_node = case.output_node
        self._parent_cond = case._parent_cond

    def _parent_conditional(self) -> Conditional:
        if self._parent_cond is None:
            msg = "If must have a parent conditional."
            raise ConditionalError(msg)
        return self._parent_cond

    @property
    def conditional_node(self) -> Node:
        """The node that represents the parent conditional."""
        return self._parent_conditional().parent_node


class If(_IfElse):
    """Build the 'if' branch of a conditional branching on a boolean value.

    Examples:
        >>> from hugr.dfg import Dfg
        >>> dfg = Dfg(tys.Qubit)
        >>> (q,) = dfg.inputs()
        >>> if_ = dfg.add_if(dfg.load(val.TRUE), q)
        >>> if_.set_outputs(if_.input_node[0])
        >>> else_= if_.add_else()
        >>> else_.set_outputs(else_.input_node[0])
        >>> dfg.hugr[else_.conditional_node].op
        Conditional(sum_ty=Bool, other_inputs=[Qubit])
    """

    def add_else(self) -> Else:
        """Finish building the 'if' branch and start building the 'else' branch."""
        return Else(self._parent_conditional().add_case(0))


class Else(_IfElse):
    """Build the 'else' branch of a conditional branching on a boolean value.

    See :class:`If` for an example.
    """

    def finish(self) -> Node:
        """Deprecated, use `conditional_node` property."""
        # TODO remove in 0.4.0
        return self.conditional_node  # pragma: no cover


@dataclass
class Conditional(ParentBuilder[ops.Conditional], AbstractContextManager):
    """Build a conditional branching on a sum type.

    Args:
        sum_ty: The sum type to branch on.
        other_inputs: The inputs for the conditional that aren't included in the
        sum variants. These are passed to all cases.

    Examples:
        >>> cond = Conditional(tys.Bool, [tys.Qubit])
        >>> cond.parent_op
        Conditional(sum_ty=Bool, other_inputs=[Qubit])
    """

    #: builders for each case and whether they have been built by the user yet
    _case_builders: list[tuple[Case, bool]] = field(default_factory=list)

    def __init__(self, sum_ty: Sum, other_inputs: TypeRow) -> None:
        root_op = ops.Conditional(sum_ty, other_inputs)
        hugr = Hugr(root_op)
        self._init_impl(hugr, hugr.root, len(sum_ty.variant_rows))

    def _init_impl(self: Conditional, hugr: Hugr, root: Node, n_cases: int) -> None:
        self.hugr = hugr
        self.parent_node = root
        self._case_builders = []

        for case_id in range(n_cases):
            new_case = Case.new_nested(
                ops.Case(self.parent_op.nth_inputs(case_id)),
                self.hugr,
                self.parent_node,
            )
            new_case._parent_cond = self
            self._case_builders.append((new_case, False))

    @property
    def cases(self) -> dict[int, Node | None]:
        """Map from case index to node holding the :class:`Case <hugr.ops.Case>`.

        DEPRECATED
        """
        # TODO remove in 0.10
        warnings.warn(
            "The 'cases' property is deprecated and"
            " will be removed in a future version.",
            DeprecationWarning,
            stacklevel=2,
        )
        return {
            i: case.parent_node if b else None
            for i, (case, b) in enumerate(self._case_builders)
        }

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args) -> None:
        if not all(built for _, built in self._case_builders):
            msg = "All cases must be added before exiting context."
            raise ConditionalError(msg)
        return None

    @classmethod
    def new_nested(
        cls,
        sum_ty: Sum,
        other_inputs: TypeRow,
        hugr: Hugr,
        parent: ToNode | None = None,
    ) -> Conditional:
        """Build a Conditional nested inside an existing HUGR graph.

        Args:
            sum_ty: The sum type to branch on.
            other_inputs: The inputs for the conditional that aren't included in the
                sum variants. These are passed to all cases.
            hugr: The HUGR instance this Conditional is part of.
            parent: The parent node for the Conditional: defaults to the root of
              the HUGR instance.

        Returns:
            The new Conditional builder.
        """
        new = cls.__new__(cls)
        root = hugr.add_node(
            ops.Conditional(sum_ty, other_inputs),
            parent or hugr.root,
        )
        new._init_impl(hugr, root, len(sum_ty.variant_rows))
        return new

    def _update_outputs(self, outputs: TypeRow) -> None:
        if self.parent_op._outputs is None:
            self.parent_op._outputs = outputs
            self.parent_node = self.hugr._update_node_outs(
                self.parent_node, len(outputs)
            )
        else:
            if outputs != self.parent_op._outputs:
                msg = "Mismatched case outputs."
                raise ConditionalError(msg)

    def add_case(self, case_id: int) -> Case:
        """Start building a case for the conditional.

        Args:
            case_id: The index of the case to build. Input types for the case
            are the corresponding variant of the sum type concatenated with the
            other inputs to the conditional.

        Returns:
            The new case builder.

        Raises:
            ConditionalError: If the case index is out of range.

        Examples:
            >>> cond = Conditional(tys.Bool, [tys.Qubit])
            >>> with cond.add_case(0) as case:\
                    case.set_outputs(*case.inputs())
        """
        if case_id >= len(self._case_builders):
            msg = f"Case {case_id} out of possible range."
            raise ConditionalError(msg)
        case, built = self._case_builders[case_id]
        if built:
            msg = f"Case {case_id} already built."
            raise ConditionalError(msg)
        self._case_builders[case_id] = (case, True)
        return case

    # TODO insert_case


@dataclass
class TailLoop(DfBase[ops.TailLoop]):
    """Builder for a tail-controlled loop.

    Args:
        just_inputs: Types that are only inputs to the loop body.
        rest: The remaining input types that are also output types.

    Examples:
        >>> tl = TailLoop([tys.Bool], [tys.Qubit])
        >>> tl.parent_op
        TailLoop(just_inputs=[Bool], rest=[Qubit])
    """

    def __init__(self, just_inputs: TypeRow, rest: TypeRow) -> None:
        root_op = ops.TailLoop(just_inputs, rest)
        super().__init__(root_op)

    def set_outputs(self, *outputs: Wire) -> None:
        super().set_outputs(*outputs)

        assert len(outputs) > 0
        sum_wire = outputs[0]
        sum_type = self.hugr.port_type(sum_wire.out_port())
        assert isinstance(sum_type, Sum)
        assert len(sum_type.variant_rows) == 2
        self._set_parent_output_count(len(sum_type.variant_rows[1]) + len(outputs) - 1)

    def set_loop_outputs(self, sum_wire: Wire, *rest: Wire) -> None:
        """Set the outputs of the loop body. The first wire must be the sum type
        that controls loop termination.

        Args:
            sum_wire: The wire holding the sum type that controls loop termination.
            rest: The remaining output wires (corresponding to the 'rest' types).
        """
        self.set_outputs(sum_wire, *rest)



"""Collection types and operations."""

from __future__ import annotations

from dataclasses import dataclass

import hugr.tys as tys
from hugr import val
from hugr.std import _load_extension
from hugr.utils import comma_sep_str

EXTENSION = _load_extension("collections.list")


@dataclass(eq=False)
class List(tys.ExtType):
    """List type with a fixed element type."""

    def __init__(self, ty: tys.Type) -> None:
        ty_arg = tys.TypeTypeArg(ty)

        self.type_def = EXTENSION.types["List"]
        self.args = [ty_arg]

    @property
    def ty(self) -> tys.Type:
        """Returns the type of the list."""
        assert isinstance(
            self.args[0], tys.TypeTypeArg
        ), "List elements must have a valid type"
        return self.args[0].ty

    def type_bound(self) -> tys.TypeBound:
        return self.ty.type_bound()


@dataclass
class ListVal(val.ExtensionValue):
    """Constant value for a list of elements."""

    v: list[val.Value]
    ty: List

    def __init__(self, v: list[val.Value], elem_ty: tys.Type) -> None:
        self.v = v
        self.ty = List(elem_ty)

    def to_value(self) -> val.Extension:
        name = "ListValue"
        # The value list must be serialized at this point, otherwise the
        # `Extension` value would not be serializable.
        vs = [v._to_serial_root() for v in self.v]
        element_ty = self.ty.ty._to_serial_root()
        serial_val = {"values": vs, "typ": element_ty}
        return val.Extension(
            name, typ=self.ty, val=serial_val, extensions=[EXTENSION.name]
        )

    def __str__(self) -> str:
        return f"[{comma_sep_str(self.v)}]"



"""Base classes for HUGR builders."""

from __future__ import annotations

from typing import (
    Protocol,
    cast,
)

from hugr.hugr.base import Hugr, OpVar
from hugr.hugr.node_port import (
    Node,
    ToNode,
)


class ParentBuilder(ToNode, Protocol[OpVar]):
    """Abstract interface implemented by builders of nodes that contain child HUGRs."""

    #: The child HUGR.
    hugr: Hugr[OpVar]
    # Unique parent node.
    parent_node: Node

    def to_node(self) -> Node:
        return self.parent_node

    @property
    def parent_op(self) -> OpVar:
        """The parent node's operation."""
        return cast(OpVar, self.hugr[self.parent_node].op)
